import os
import glob
import pickle
import pytest
import numpy as np
import astropy.io.fits as fits
import corgidrp
from corgidrp.data import Image, Dataset
from corgidrp.mocks import create_default_L1_headers, create_dark_calib_files

np.random.seed(123)

data = np.ones([1024,1024]) * 2
err = np.zeros([1024,1024])
err1 = np.ones([1024,1024])
err2 = err1.copy()
err3 = np.ones([1,1024,1024]) * 0.5
dq = np.zeros([1024,1024], dtype = int)
dq1 = dq.copy()
dq1[0,0] = 1
prhd, exthd = create_default_L1_headers()
errhd = fits.Header()
errhd["CASE"] = "test"
dqhd = fits.Header()
dqhd["CASE"] = "test"


def test_hashing():
    """
    Test hashing works on data, err, and dq at the same time
    Two images with same data should be the same
    """
    # identical images should get the same hash
    image1 = Image(data, err = err, dq = dq, pri_hdr = prhd, ext_hdr = exthd)
    image2 = Image(np.copy(data), err = np.copy(err), dq = np.copy(dq), pri_hdr = prhd, ext_hdr = exthd)

    assert image1.get_hash() == image2.get_hash()

    # modifying the data should result in different hashes
    image2.data += 1

    assert image1.get_hash() != image2.get_hash()

    # take image 2 and modify the error. should be different hash from before
    old_hash = image2.get_hash()

    image2.err += 1
    assert old_hash != image2.get_hash()

    # take image 2 and modify the dq frame. should be different hash from before
    old_hash = image2.get_hash()

    image2.dq[0] = 1
    assert old_hash != image2.get_hash()

def test_split_dataset():
    """
    Test splitting dataset into sub datasets based on header keywords
    """
    image1 = Image(np.copy(data), err = np.copy(err), dq = np.copy(dq), pri_hdr = prhd.copy(), ext_hdr = exthd.copy())
    image2 = Image(np.copy(data), err = np.copy(err), dq = np.copy(dq), pri_hdr = prhd.copy(), ext_hdr = exthd.copy())
    image3 = Image(np.copy(data), err = np.copy(err), dq = np.copy(dq), pri_hdr = prhd.copy(), ext_hdr = exthd.copy())
    image4 = Image(np.copy(data), err = np.copy(err), dq = np.copy(dq), pri_hdr = prhd.copy(), ext_hdr = exthd.copy())

    orig_dataset = Dataset([image1, image2, image3, image4])

    # defaults
    # exthdr['EXPTIME'] = 60.0
    # prihdr['OBSID'] = 0

    ## slice it into 2
    image1.pri_hdr['OBSNUM'] = 0
    image1.ext_hdr['EXPTIME'] = 60.0

    image2.pri_hdr['OBSNUM'] = 1
    image2.ext_hdr['EXPTIME'] = 120.

    image3.pri_hdr['OBSNUM'] = 0
    image3.ext_hdr['EXPTIME'] = 60.0

    image4.pri_hdr['OBSNUM'] = 1
    image4.ext_hdr['EXPTIME'] = 120.

    sliced_datasets, unique_combos = orig_dataset.split_dataset(exthdr_keywords=['EXPTIME',], prihdr_keywords=['OBSNUM',])
    assert len(sliced_datasets) == 2

    sliced_datasets, unique_combos = orig_dataset.split_dataset(exthdr_keywords=['EXPTIME',])
    assert len(sliced_datasets) == 2

    sliced_datasets, unique_combos = orig_dataset.split_dataset(prihdr_keywords=['OBSNUM',])
    assert len(sliced_datasets) == 2

    ## slice it into 3
    image1.pri_hdr['OBSNUM'] = 0
    image1.ext_hdr['EXPTIME'] = 60.0

    image2.pri_hdr['OBSNUM'] = 1
    image2.ext_hdr['EXPTIME'] = 60.0

    image3.pri_hdr['OBSNUM'] = 0
    image3.ext_hdr['EXPTIME'] = 60.0

    image4.pri_hdr['OBSNUM'] = 1
    image4.ext_hdr['EXPTIME'] = 120.


    sliced_datasets, unique_combos = orig_dataset.split_dataset(exthdr_keywords=['EXPTIME',], prihdr_keywords=['OBSNUM',])
    assert len(sliced_datasets) == 3

    ## slice it into 4
    image1.pri_hdr['OBSNUM'] = 0
    image1.ext_hdr['EXPTIME'] = 60.0

    image2.pri_hdr['OBSNUM'] = 1
    image2.ext_hdr['EXPTIME'] = 60.0

    image3.pri_hdr['OBSNUM'] = 0
    image3.ext_hdr['EXPTIME'] = 120.

    image4.pri_hdr['OBSNUM'] = 1
    image4.ext_hdr['EXPTIME'] = 120.

    sliced_datasets, unique_combos = orig_dataset.split_dataset(exthdr_keywords=['EXPTIME',], prihdr_keywords=['OBSNUM',])
    assert len(sliced_datasets) == 4

    sliced_datasets, unique_combos = orig_dataset.split_dataset(exthdr_keywords=['EXPTIME',])
    assert len(sliced_datasets) == 2



def test_pickling():
    """
    Test that datasets and images can be pickled
    """
    ###### create simulated data
    # check that simulated data folder exists, and create if not
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)

    create_dark_calib_files(filedir=datadir)

    ####### test data architecture
    dark_filenames = glob.glob(os.path.join(datadir, "simcal_dark*.fits"))

    dark_dataset = Dataset(dark_filenames)

    pickle_filename = os.path.join(datadir, "simcal_dataset.pkl")
    pickled = pickle.dumps(dark_dataset)
    pickled_dark_dataset = pickle.loads(pickled)

    assert np.all(dark_dataset[0].data == pickled_dark_dataset[0].data)


if __name__ == "__main__":
    test_hashing()
    test_split_dataset()
    test_pickling()


"""
Test the walker infrastructure to read and execute recipes
"""
import os
import glob
import json
import numpy as np
import astropy.time as time
import astropy.io.fits as fits
import corgidrp
import corgidrp.data as data
import corgidrp.mocks as mocks
import corgidrp.caldb as caldb
import corgidrp.walker as walker
import corgidrp.detector as detector

np.random.seed(456)

def test_autoreducing():
    """
    Tests both generating and processing a basic L1->L2a SCI recipe
    """
    # create dirs
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "walker_output")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)


    # create simulated data
    l1_dataset = mocks.create_prescan_files(filedir=datadir, arrtype="SCI", numfiles=2)
    # simulate the expected CGI naming convention
    fname_template = "CGI_0200001999001000{:03d}_20250415T0305102_L1_.fits"
    for i, image in enumerate(l1_dataset):
        image.filename = fname_template.format(i)
    l1_dataset.save(filedir=datadir)
    filelist = [frame.filepath for frame in l1_dataset]


    ###########################################
    ### Create a dummy non-linearity file ####
    #Create a mock dataset because it is a required input when creating a NonLinearityCalibration
    dummy_dataset = mocks.create_prescan_files()

    # Make a non-linearity correction calibration file
    input_non_linearity_filename = "nonlin_table_TVAC.txt"
    input_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", input_non_linearity_filename)
    test_non_linearity_filename = input_non_linearity_filename.split(".")[0] + ".fits"
    nonlin_fits_filepath = os.path.join(os.path.dirname(__file__), "test_data", test_non_linearity_filename)
    tvac_nonlin_data = np.genfromtxt(input_non_linearity_path, delimiter=",")

    pri_hdr, ext_hdr = mocks.create_default_L1_headers()
    new_nonlinearity = data.NonLinearityCalibration(tvac_nonlin_data,pri_hdr=pri_hdr,ext_hdr=ext_hdr,input_dataset = dummy_dataset)
    new_nonlinearity.filename = nonlin_fits_filepath
    new_nonlinearity.save()
    # index the sample nonlinearity correction that we need for processing
    # fake the headers because this frame doesn't have the proper headers
    prihdr, exthdr = mocks.create_default_L1_headers("SCI")
    new_nonlinearity.pri_hdr = prihdr
    new_nonlinearity.ext_hdr = exthdr
    new_nonlinearity.ext_hdr.set('DRPCTIME', time.Time.now().isot, "When this file was saved")
    new_nonlinearity.ext_hdr.set('DRPVERSN', corgidrp.__version__, "corgidrp version that produced this file")
    mycaldb = caldb.CalDB()
    mycaldb.create_entry(new_nonlinearity)
    
    #Make a KGain calibration file
    kgain_arr = np.array([[8.8]])
    new_kgain = data.KGain(kgain_arr,pri_hdr=prihdr,ext_hdr=exthdr,input_dataset = dummy_dataset)
    new_kgain.ext_hdr.set('DRPCTIME', time.Time.now().isot, "When this file was saved")
    new_kgain.ext_hdr.set('DRPVERSN', corgidrp.__version__, "corgidrp version that produced this file")
    new_kgain.save(filedir = os.path.join(os.path.dirname(__file__), "test_data"), filename = "kgain.fits")
    
    mycaldb.create_entry(new_kgain)
    

    CPGS_XML_filepath = "" # not yet implemented

    # generate recipe and run it
    # basic l2a recipe to keep things simple
    recipe = walker.walk_corgidrp(filelist, CPGS_XML_filepath, outputdir, template="l1_to_l2a_basic.json")

    # check that the output dataset is saved to the output dir
    # filenames have been updated to L2a. 
    output_files = [os.path.join(outputdir, frame.filename.replace("_L1_", "_L2a")) for frame in l1_dataset]
    output_dataset = data.Dataset(output_files)
    assert len(output_dataset) == len(l1_dataset) # check the same number of files
    # check that the recipe is saved into the header.

    for frame in output_dataset:
        assert "RECIPE" in frame.ext_hdr
        # test recipe was correctly written into the header
        # do a string comparison, easiest way to check
        hdr_recipe = json.loads(frame.ext_hdr["RECIPE"])
        assert json.dumps(hdr_recipe) == json.dumps(recipe)

    # clean up
    mycaldb.remove_entry(new_nonlinearity)
    mycaldb.remove_entry(new_kgain)

def test_auto_template_identification():
    """
    Tests the process of coming up with the right template and filling in calibration files
    """
    # create dirs
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "walker_output")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)

    # create simulated data
    l1_dataset = mocks.create_prescan_files(filedir=datadir, arrtype="SCI", numfiles=2)
    # simulate the expected CGI naming convention
    fname_template = "CGI_0200001999001000{:03d}_20250415T0305102_L1_.fits"
    for i, image in enumerate(l1_dataset):
        image.filename = fname_template.format(i)
    l1_dataset.save(filedir=datadir)
    filelist = [frame.filepath for frame in l1_dataset]


    ###### Setup necessary calibration files
    # Create necessary calibration files
    # we are going to make calibration files using
    # a combination of the II&T nonlinearty file and the mock headers from
    # our unit test version
    pri_hdr, ext_hdr = mocks.create_default_calibration_product_headers()
    ext_hdr["DRPCTIME"] = time.Time.now().isot
    ext_hdr['DRPVERSN'] =  corgidrp.__version__
    mock_input_dataset = data.Dataset(filelist)

    this_caldb = caldb.CalDB() # connection to cal DB

    # Nonlinearity calibration
    ### Create a dummy non-linearity file ####
    #Create a mock dataset because it is a required input when creating a NonLinearityCalibration
    dummy_dataset = mocks.create_prescan_files()

    # Make a non-linearity correction calibration file
    input_non_linearity_filename = "nonlin_table_TVAC.txt"
    input_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", input_non_linearity_filename)
    test_non_linearity_filename = input_non_linearity_filename.split(".")[0] + ".fits"
    nonlin_fits_filepath = os.path.join(os.path.dirname(__file__), "test_data", test_non_linearity_filename)
    tvac_nonlin_data = np.genfromtxt(input_non_linearity_path, delimiter=",")

    pri_hdr, ext_hdr = mocks.create_default_calibration_product_headers()
    new_nonlinearity = data.NonLinearityCalibration(tvac_nonlin_data,pri_hdr=pri_hdr,ext_hdr=ext_hdr,input_dataset = dummy_dataset)
    new_nonlinearity.filename = nonlin_fits_filepath
    # fake the headers because this frame doesn't have the proper headers
    new_nonlinearity.pri_hdr = pri_hdr
    new_nonlinearity.ext_hdr = ext_hdr
    this_caldb.create_entry(new_nonlinearity)

    # KGain
    kgain_val = 8.7
    kgain = data.KGain(np.array([[kgain_val]]), pri_hdr=pri_hdr, ext_hdr=ext_hdr, 
                    input_dataset=mock_input_dataset)
    kgain.save(filedir=outputdir, filename="mock_kgain.fits")
    this_caldb.create_entry(kgain)

    # NoiseMap
    noise_map_dat = np.zeros((3, detector.detector_areas['SCI']['frame_rows'], detector.detector_areas['SCI']['frame_cols']))
    rows, cols, r0c0 = detector.unpack_geom('SCI', 'image')
    noise_map_noise = np.zeros([1,] + list(noise_map_dat.shape))
    noise_map_dq = np.zeros(noise_map_dat.shape, dtype=int)
    err_hdr = fits.Header()
    err_hdr['BUNIT'] = 'detected electrons'
    ext_hdr['B_O'] = 0
    ext_hdr['B_O_ERR'] = 0
    noise_map = data.DetectorNoiseMaps(noise_map_dat, pri_hdr=pri_hdr, ext_hdr=ext_hdr,
                                    input_dataset=mock_input_dataset, err=noise_map_noise,
                                    dq = noise_map_dq, err_hdr=err_hdr)
    noise_map.save(filedir=outputdir, filename="mock_detnoisemaps.fits")
    this_caldb.create_entry(noise_map)

    ## Flat field
    flat_dat = np.ones((1024, 1024))
    flat = data.FlatField(flat_dat, pri_hdr=pri_hdr, ext_hdr=ext_hdr, input_dataset=mock_input_dataset)
    flat.save(filedir=outputdir, filename="mock_flat.fits")
    this_caldb.create_entry(flat)

    # bad pixel map
    bp_dat = np.zeros((1024, 1024), dtype=int)
    bp_map = data.BadPixelMap(bp_dat, pri_hdr=pri_hdr, ext_hdr=ext_hdr, input_dataset=mock_input_dataset)
    bp_map.save(filedir=outputdir, filename="mock_bpmap.fits")
    this_caldb.create_entry(bp_map)


    ### Finally, test the recipe identification
    recipe = walker.autogen_recipe(filelist, outputdir)

    assert recipe['name'] == 'l1_to_l2a_basic'
    assert recipe['template'] == False

    # now cleanup
    this_caldb.remove_entry(new_nonlinearity)
    this_caldb.remove_entry(kgain)
    this_caldb.remove_entry(noise_map)
    this_caldb.remove_entry(flat)
    this_caldb.remove_entry(bp_map)


def test_saving():
    """
    Tests the special save function including suffix. Tries both calibration image and non-calibration image
    """
    ### create dirs
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "walker_output")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)

    ### load test recipe
    testdatadir = os.path.join(os.path.dirname(__file__), "test_data")
    save_recipe_path = os.path.join(testdatadir, "saving_only.json")
    save_recipe = json.load(open(save_recipe_path, 'r'))
    
    ######################
    ## Test regular images
    ######################

    ### Create mock Image data
    l1_dataset = mocks.create_dark_calib_files(filedir=datadir, numfiles=2)
    # simulate the expected CGI naming convention
    fname_template = "CGI_0200001999001000{:03d}_20250415T0305102_L1_.fits"
    for i, image in enumerate(l1_dataset):
        image.filename = fname_template.format(i)
    l1_dataset.save(filedir=datadir)
    filelist = [frame.filepath for frame in l1_dataset]

    this_recipe = walker.autogen_recipe(filelist, outputdir, template=save_recipe)
    walker.run_recipe(this_recipe)

    # check that the output dataset is saved to the output dir
    # filenames have been appended with a suffix
    output_files = [os.path.join(outputdir, "CGI_0200001999001000{:03d}_20250415T0305102_L1__test.fits".format(i)) for i in range(len(l1_dataset))]
    output_dataset = data.Dataset(output_files)
    assert len(output_dataset) == len(l1_dataset) # check the same number of files
    
    ##########################
    ## Test calibration image 
    ##########################
    # Fake a nonlinearity dataset
    ### Create a dummy non-linearity file ####
    #Create a mock dataset because it is a required input when creating a NonLinearityCalibration
    dummy_dataset = mocks.create_prescan_files()

    # Make a non-linearity correction calibration file
    input_non_linearity_filename = "nonlin_table_TVAC.txt"
    input_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", input_non_linearity_filename)
    tvac_nonlin_data = np.genfromtxt(input_non_linearity_path, delimiter=",")
    test_non_linearity_filename = input_non_linearity_filename.split(".")[0] + ".fits"
    nonlin_fits_filepath = os.path.join(os.path.dirname(__file__), "test_data", test_non_linearity_filename)
    pri_hdr, ext_hdr = mocks.create_default_calibration_product_headers()
    fake_nonlinearity = data.NonLinearityCalibration(tvac_nonlin_data,pri_hdr=pri_hdr,ext_hdr=ext_hdr,input_dataset = dummy_dataset)
    fake_nonlinearity.filename = nonlin_fits_filepath
    # fake the headers because this frame doesn't have the proper headers
    prihdr, exthdr = mocks.create_default_L1_headers("SCI")
    fake_nonlinearity.pri_hdr = prihdr
    fake_nonlinearity.ext_hdr = exthdr
    fake_nonlinearity.ext_hdr['DATATYPE'] = 'NonLinearityCalibration'
    fake_nonlinearity.ext_hdr.set('DRPCTIME', time.Time.now().isot, "When this file was saved")
    fake_nonlinearity.ext_hdr.set('DRPVERSN', corgidrp.__version__, "corgidrp version that produced this file")
    fake_nonlinearity.filename = "CGI_test.fits"

    # tested the run_recipe portion of the code already (nothing different)
    # test save_data when we only pass in a single calibration file and not a dataset to see how it goes
    walker.save_data(fake_nonlinearity, outputdir, "nonlin")
    output_filepath = os.path.join(outputdir, "CGI_test_nonlin.fits")
    new_nonlinearity = data.NonLinearityCalibration(output_filepath)

    # remove this entry. should only work if it's already in the database
    # also used for cleanup
    this_caldb = caldb.CalDB()
    this_caldb.remove_entry(new_nonlinearity)



def test_skip_missing_calib():
    """
    Tests the option of skipping steps with missing calibrations
    """
    # turn on skipping
    old_setting = corgidrp.skip_missing_cal_steps
    corgidrp.skip_missing_cal_steps = True

    # use an empty test caldb
    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    if not os.path.exists(calibdir):
        os.mkdir(calibdir)
    testcaldb_filepath = os.path.join(calibdir, "empty_caldb.csv")
    old_caldb_filepath = corgidrp.caldb_filepath
    corgidrp.caldb_filepath = testcaldb_filepath

    # create dirs
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "walker_output")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)

    # create simulated data
    l1_dataset = mocks.create_prescan_files(filedir=datadir, arrtype="SCI", numfiles=2)
    # simulate the expected CGI naming convention
    fname_template = "CGI_0200001999001000{:03d}_20250415T0305102_L1_.fits"
    for i, image in enumerate(l1_dataset):
        image.filename = fname_template.format(i)
    l1_dataset.save(filedir=datadir)
    filelist = [frame.filepath for frame in l1_dataset]


    ### Test that we are skipping the steps without calibrations
    # use l1 to l2b recipe
    template_filepath = os.path.join(os.path.dirname(walker.__file__), "recipe_templates", "l1_to_l2b.json")
    template_recipe = json.load(open(template_filepath, "r"))
    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)

    assert recipe['name'] == 'l1_to_l2b'
    assert recipe['template'] == False

    assert recipe['steps'][0]['skip'] # prescan bias sub    
    assert recipe['steps'][2]['skip'] # nonlinearity
    assert recipe['steps'][6]['skip'] # kgain
    
    # cut down to recipe to just the first 5 steps (to the first save)
    recipe['steps'] = recipe['steps'][:5]

    # run with all the ksips
    walker.run_recipe(recipe, save_recipe_file=False)

    # check that the output dataset is saved to the output dir
    # filenames have been appended with a suffix
    output_files = [os.path.join(outputdir, "CGI_0200001999001000{:03d}_20250415T0305102_L2a.fits".format(i)) for i in range(len(l1_dataset))]
    output_dataset = data.Dataset(output_files)
    assert len(output_dataset) == len(l1_dataset) # check the same number of files

    for hist_entry in output_dataset[0].ext_hdr['HISTORY']:
        assert 'non-linearity' not in hist_entry.lower()

    corgidrp.skip_missing_cal_steps = old_setting
    corgidrp.caldb_filepath = old_caldb_filepath



def test_skip_missing_optional_calib():
    """
    Tests optional calibrtion behavior when skpip_missing_calibs is True
    The behavior is that the step should not be skipped, given the calibration is optional
    """
    # turn on skipping
    old_setting = corgidrp.skip_missing_cal_steps
    corgidrp.skip_missing_cal_steps = True

    # use an empty test caldb
    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    if not os.path.exists(calibdir):
        os.mkdir(calibdir)
    testcaldb_filepath = os.path.join(calibdir, "empty_caldb.csv")
    old_caldb_filepath = corgidrp.caldb_filepath
    corgidrp.caldb_filepath = testcaldb_filepath

    # create dirs
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "walker_output")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)

    # create simulated data
    l1_dataset = mocks.create_prescan_files(filedir=datadir, arrtype="SCI", numfiles=2)
    # simulate the expected CGI naming convention
    fname_template = "CGI_0200001999001000{:03d}_20250415T0305102_L1_.fits"
    for i, image in enumerate(l1_dataset):
        image.filename = fname_template.format(i)
    l1_dataset.save(filedir=datadir)
    filelist = [frame.filepath for frame in l1_dataset]


    # use l1 to l2a recipe since it as optional calibration
    template_filepath = os.path.join(os.path.dirname(walker.__file__), "recipe_templates", "l1_to_l2a_basic.json")
    template_recipe = json.load(open(template_filepath, "r"))

    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)

    # check prescan bias sub is not skipped and Detector Noise Maps is None
    assert 'skip' not in recipe['steps'][0] # prescan biassub
    assert recipe['steps'][0]['calibs']['DetectorNoiseMaps'] is None

    # assert nonlinearity is indeed skipped
    assert recipe['steps'][2]['skip'] # nonlinearity

    corgidrp.skip_missing_cal_steps = old_setting
    corgidrp.caldb_filepath = old_caldb_filepath

def test_jit_calibs():
    """
    Tests defining calibrations just in time
    """
    old_setting = corgidrp.jit_calib_id
    corgidrp.jit_calib_id = True

    # create dirs
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "walker_output")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)


    # create simulated data
    l1_dataset = mocks.create_prescan_files(filedir=datadir, arrtype="SCI", numfiles=2)
    # simulate the expected CGI naming convention
    fname_template = "CGI_0200001999001000{:03d}_20250415T0305102_L1_.fits"
    for i, image in enumerate(l1_dataset):
        image.filename = fname_template.format(i)
    l1_dataset.save(filedir=datadir)
    filelist = [frame.filepath for frame in l1_dataset]


    # index the sample nonlinearity correction that we need for processing
    # Fake a nonlinearity dataset
    ### Create a dummy non-linearity file ####
    #Create a mock dataset because it is a required input when creating a NonLinearityCalibration
    dummy_dataset = mocks.create_prescan_files()

    # Make a non-linearity correction calibration file
    input_non_linearity_filename = "nonlin_table_TVAC.txt"
    input_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", input_non_linearity_filename)
    tvac_nonlin_data = np.genfromtxt(input_non_linearity_path, delimiter=",")
    test_non_linearity_filename = input_non_linearity_filename.split(".")[0] + ".fits"
    nonlin_fits_filepath = os.path.join(os.path.dirname(__file__), "test_data", test_non_linearity_filename)
    pri_hdr, ext_hdr = mocks.create_default_calibration_product_headers()
    new_nonlinearity = data.NonLinearityCalibration(tvac_nonlin_data,pri_hdr=pri_hdr,ext_hdr=ext_hdr,input_dataset = dummy_dataset)
    new_nonlinearity.filename = nonlin_fits_filepath
    # fake the headers because this frame doesn't have the proper headers
    prihdr, exthdr = mocks.create_default_L1_headers("SCI")
    new_nonlinearity.pri_hdr = prihdr
    new_nonlinearity.ext_hdr = exthdr
    new_nonlinearity.ext_hdr.set('DRPCTIME', time.Time.now().isot, "When this file was saved")
    new_nonlinearity.ext_hdr.set('DRPVERSN', corgidrp.__version__, "corgidrp version that produced this file")
    mycaldb = caldb.CalDB()
    mycaldb.create_entry(new_nonlinearity)

    #Make a KGain calibration file
    kgain_arr = np.array([[8.8]])
    new_kgain = data.KGain(kgain_arr,pri_hdr=prihdr,ext_hdr=exthdr,input_dataset = dummy_dataset)
    new_kgain.ext_hdr.set('DRPCTIME', time.Time.now().isot, "When this file was saved")
    new_kgain.ext_hdr.set('DRPVERSN', corgidrp.__version__, "corgidrp version that produced this file")
    new_kgain.save(filedir = os.path.join(os.path.dirname(__file__), "test_data"), filename = "kgain.fits")
    
    mycaldb.create_entry(new_kgain)

    CPGS_XML_filepath = "" # not yet implemented

    # generate recipe and check we haven't defined anyhting yet
    template_filepath = os.path.join(os.path.dirname(walker.__file__), "recipe_templates", "l1_to_l2a_basic.json")
    template_recipe = json.load(open(template_filepath, "r"))
    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)

    assert recipe['steps'][2]['calibs']['NonLinearityCalibration'] == 'AUTOMATIC' 

    walker.run_recipe(recipe)

    # check that the output dataset is saved to the output dir
    # filenames have been updated to L2a. 
    output_files = [os.path.join(outputdir, frame.filename.replace("_L1_", "_L2a")) for frame in l1_dataset]
    output_dataset = data.Dataset(output_files)
    assert len(output_dataset) == len(l1_dataset) # check the same number of files
    
    # check that the recipe is saved into the header with specified calibrations
    new_recipe = json.loads(output_dataset[0].ext_hdr['RECIPE'])
    assert recipe['steps'][2]['calibs']['NonLinearityCalibration'] != 'AUTOMATIC' 


    #### Test cases where JIT should be enabled or not
    # already tested pipeline setting True, nothing set in recipe. Resulted in keeping automatic keyword

    # pipeline setting false, and nothing set in recipe. Should define calibrations
    corgidrp.jit_calib_id = False
    template_recipe = json.load(open(template_filepath, "r"))
    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)
    assert recipe['steps'][2]['calibs']['NonLinearityCalibration'] != 'AUTOMATIC' 

    # pipeline setting false, but recipe says JIT. Should keep automatic calibration
    corgidrp.jit_calib_id = False
    template_recipe = json.load(open(template_filepath, "r"))
    template_recipe['drpconfig']['jit_calib_id'] = True
    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)
    assert recipe['steps'][2]['calibs']['NonLinearityCalibration'] == 'AUTOMATIC' 

    # pipeline setting false, and recipe says no JIT. Should define calibrations
    corgidrp.jit_calib_id = False
    template_recipe = json.load(open(template_filepath, "r"))
    template_recipe['drpconfig']['jit_calib_id'] = False
    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)
    assert recipe['steps'][2]['calibs']['NonLinearityCalibration'] != 'AUTOMATIC' 

    # pipeline setting True, and recipe says no JIT. Should define calibrations
    corgidrp.jit_calib_id = True
    template_recipe = json.load(open(template_filepath, "r"))
    template_recipe['drpconfig']['jit_calib_id'] = False
    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)
    assert recipe['steps'][2]['calibs']['NonLinearityCalibration'] != 'AUTOMATIC' 

    # pipeline setting True, and recipe says JIT. Should keep automatic calibration
    corgidrp.jit_calib_id = True
    template_recipe = json.load(open(template_filepath, "r"))
    template_recipe['drpconfig']['jit_calib_id'] = True
    recipe = walker.autogen_recipe(filelist, outputdir, template=template_recipe)
    assert recipe['steps'][2]['calibs']['NonLinearityCalibration'] == 'AUTOMATIC' 


    # clean up
    mycaldb.remove_entry(new_nonlinearity)
    mycaldb.remove_entry(new_kgain)

    corgidrp.jit_calib_id = old_setting



def test_generate_multiple_recipes():
    """
    Tests that we can generate multiple recipes when passing in a dataset
    """
    # create dirs
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "walker_output")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)
    # Make a non-linearity correction calibration file
    input_non_linearity_filename = "nonlin_table_TVAC.txt"
    test_non_linearity_filename = input_non_linearity_filename.split(".")[0] + ".fits"
    test_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", test_non_linearity_filename)

    dataset = mocks.create_nonlinear_dataset(test_non_linearity_path, filedir=datadir)
    # add vistype
    for frame in dataset:
        frame.pri_hdr['VISTYPE'] = "PUPILIMG"
    dataset.save()
    filelist = [frame.filepath for frame in dataset]

    recipes = walker.autogen_recipe(filelist, outputdir)

    assert len(recipes) == 2




if __name__ == "__main__":#
    test_autoreducing()
    test_auto_template_identification()
    test_saving()
    test_skip_missing_calib()
    test_skip_missing_optional_calib()
    test_jit_calibs()
    test_generate_multiple_recipes()






from setuptools import setup, find_packages



def get_requires():
    reqs = []
    for line in open("requirements.txt", "r").readlines():
        reqs.append(line)
    return reqs

setup(
    name='corgidrp',
    version="2.0",
    description='(Roman Space Telescope) CORonaGraph Instrument Data Reduction Pipeline',
    #long_description="",
    #long_description_content_type="text/markdown",
    url='https://github.com/roman-corgi/corgidrp',
    author='Roman Coronagraph Instrument CPP',
    #author_email='',
    license='BSD',
    packages=find_packages(),
    classifiers=[
        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',

        'Programming Language :: Python :: 3',
        ],
    keywords='Roman Space Telescope Exoplanets Astronomy',
    install_requires=get_requires()
    )




"""
Module to support frame combination
"""
import numpy as np
import corgidrp.data as data


def combine_images(data_subset, err_subset, dq_subset, collapse, num_frames_scaling):
    """
    Combines several images together

    Args:
        data_subset (np.array): 3-D array of N 2-D images
        err_subset (np.array): 4-D array of N 2-D error maps
        dq_subset (np.array): 3-D array of N 2-D DQ maps
        collapse (str): "mean" or "median". 
        num_frames_scaling (bool): Multiply by number of frames in sequence in order to ~conserve photons

    Returns:
        np.array: 2-D array of combined images
        np.array: 3-D array of combined error map
        np.array: 2-D array of combined DQ maps
    """
    tot_frames = data_subset.shape[0]
    # mask bad pixels
    bad = np.where(dq_subset > 0)
    data_subset[bad] = np.nan
    err_subset[bad[0],:,bad[1],bad[2]] = np.nan
    # track the number of good values that go into the combination
    n_samples = np.ones(data_subset.shape)
    n_samples[bad] = 0
    n_samples = np.sum(n_samples, axis=0)
    if collapse.lower() == "mean":
        data_collapse = np.nanmean(data_subset, axis=0)
        err_collapse = np.sqrt(np.nanmean(err_subset**2, axis=0)) /np.sqrt(n_samples) # correct assuming standard error propagation
    elif collapse.lower() == "median":
        data_collapse = np.nanmedian(data_subset, axis=0)
        err_collapse = np.sqrt(np.nanmean(err_subset**2, axis=0)) /np.sqrt(n_samples) * np.sqrt(np.pi/2) # inflate median error
    if num_frames_scaling:
        # scale up by the number of frames
        data_collapse *= tot_frames
        err_collapse *= tot_frames
    
    # dq collpase: keep all flags on
    dq_collapse = np.bitwise_or.reduce(dq_subset, axis=0)
    # except for those pixels that have been replaced with good values
    dq_collapse[np.where((dq_collapse > 0) & (~np.isnan(data_collapse)))] = 0

    return data_collapse, err_collapse, dq_collapse



def combine_subexposures(input_dataset, num_frames_per_group=None, collapse="mean", num_frames_scaling=True):
    """
    Combines a sequence of exposures assuming a constant nubmer of frames per group. 
    The length of the dataset must be divisible by the number of frames per group.
    
    The combination is done with either the mean or median, but the collapsed image can be scaled 
    in order to ~conserve the total number of photons in the input dataset (this essentially turns a
    median into a sum)
    
    Args:
        input_dataset (corgidrp.data.Dataset): input data. 
        num_frames_per_group (int): number of subexposures per group. If None, combines all images together
        collapse (str): "mean" or "median". (default: mean) 
        num_frames_scaling (bool): Multiply by number of frames in sequence in order to ~conserve photons (default: True)

    Returns:
        corgidrp.data.Dataset: dataset after combination of every "num_frames_per_group" frames together
    """
    if num_frames_per_group is None:
        num_frames_per_group = len(input_dataset)

    if len(input_dataset) % num_frames_per_group != 0:
        raise ValueError("Input dataset of length {0} cannot be grouped in sets of {1}".format(len(input_dataset), num_frames_per_group))
    
    if collapse.lower() not in ["mean", "median"]:
        raise ValueError("combine_subexposures can only collapse with mean or median")

    num_groups = len(input_dataset) // num_frames_per_group
    new_dataset = []
    for i in range(num_groups):
        data_subset = np.copy(input_dataset.all_data[num_frames_per_group*i:num_frames_per_group*(i+1)])
        err_subset = np.copy(input_dataset.all_err[num_frames_per_group*i:num_frames_per_group*(i+1)])
        dq_subset = np.copy(input_dataset.all_dq[num_frames_per_group*i:num_frames_per_group*(i+1)])

        data_collapse, err_collapse, dq_collapse = combine_images(data_subset, err_subset, dq_subset, collapse=collapse, 
                                                                  num_frames_scaling=num_frames_scaling)

        # grab the headers from the first frame in this sub sequence
        pri_hdr = input_dataset[num_frames_per_group*i].pri_hdr.copy()
        ext_hdr = input_dataset[num_frames_per_group*i].ext_hdr.copy()
        ext_hdr["NUM_FR"] = num_frames_per_group
        err_hdr = input_dataset[num_frames_per_group*i].err_hdr.copy()
        dq_hdr = input_dataset[num_frames_per_group*i].dq_hdr.copy()
        hdulist = input_dataset[num_frames_per_group*i].hdu_list.copy()
        new_image = data.Image(data_collapse, pri_hdr=pri_hdr, ext_hdr=ext_hdr, err=err_collapse, dq=dq_collapse, err_hdr=err_hdr, 
                                dq_hdr=dq_hdr, input_hdulist=hdulist)
                                
        # always take the last filename in the group for the combined frame
        last_idx_in_group = num_frames_per_group*(i+1) - 1
        new_image.filename = input_dataset[last_idx_in_group].filename   

        new_image._record_parent_filenames(input_dataset[num_frames_per_group*i:num_frames_per_group*(i+1)])   
        new_dataset.append(new_image)
    new_dataset = data.Dataset(new_dataset)
    new_dataset.update_after_processing_step("Combine_subexposures: combined every {0} frames by {1}".format(num_frames_per_group, collapse))

    return new_dataset


import os
import numpy as np
from astropy.table import Table
import corgidrp.fluxcal as fluxcal
import corgidrp.klip_fm as klip_fm
from corgidrp.data import Image, Dataset
import corgidrp.l4_to_tda as l4_to_tda
import corgidrp.l3_to_l4 as l3_to_l4
import matplotlib.pyplot as plt
from astropy.visualization import ZScaleInterval, ImageNormalize
import pyklip.fakes


def measure_companions(
    host_star_image,
    psf_sub_image,
    ct_cal,
    fpam_fsam_cal,
    nd_cal=None,
    phot_method='aperture',
    photometry_kwargs=None,
    fluxcal_factor=None,
    host_star_in_calspec=True,
    thrp_corr="L4",
    coronagraphic_dataset=None,
    refstar_dataset=None,
    klip_fm_kwargs=None,
    verbose=True,
    kl_mode_idx=-1,
    cand_locs=None
):
    """
    Measure companion properties in a coronagraphic image and return a table with companion position, flux ratio, and apparent magnitude.
    
    Args:
        host_star_image (corgidrp.data.Image): Unocculted 2-D image of the host star (potentially corrected for ND transmission)
        psf_sub_image (corgidrp.data.Image): PSF-subtracted image with companions.
        ct_cal (corgidrp.data.CoreThroughputCalibration): Core throughput calibration data.
        fpam_fsam_cal (corgidrp.data.FpamFsamCal): Transformation calibration data.
        nd_cal (corgidrp.data.NDFilterSweetSpotDataset): ND calibration if host_star_image has not been corrected for ND. If None,
            no correction will be done.
        phot_method (str): Photometry method to use ('aperture' or 'gauss2d').
        photometry_kwargs (dict): Dictionary of keyword arguments for photometry.
        fluxcal_factor (corgidrp.Data.FluxcalFactor): Flux calibration factor object.
        host_star_in_calspec (bool): Flag indicating whether to use host star magnitude from calspec.
        thrp_corr (str): How to do the algorithm throughput estimation to measure the flux. Options are:
                         1) "L4", which uses the algorithm throughput calculated from L3 -> L4 processing
                         2) "None", which ignorees algorithm throughput
                         3) "KLIP-FM", which forward models the PSF through KLIP (requires passing in 
                         coronagraphic_dataset and refstar_dataset, if RDI). [not implemented]
        coronagraphic_dataset (corgidrp.data.Dataset): Dataset containing coronagraphic images.
        refstar_dataset (corgidrp.data.Dataset): RDI reference star dataset for PSF subtraction
        klip_fm_kwargs (dict): keyword arguments to pass into forward_model_psf()
        verbose (bool): Flag to enable verbose output.
        kl_mode_idx (int): Index of the KL mode to use (must match the one used for PSF subtraction).
            Defaults to the last KL mode.
        cand_locs (list of tuples, optional): Locations of known off-axis sources to measure flux. 
            Each tuple should be of the format (sep_pix, pa_degrees). If not, the function
            will look in the header for detectios from l4_to_tda.find_source 
    
    Returns:
        result_table (astropy.table.Table): Table containing companion measurements.
    """
    # Measure counts of the host star image
    guess_index = np.unravel_index(np.nanargmax(host_star_image.data), host_star_image.data.shape)
    host_star_peakflux, host_star_fwhm, x_host, y_host = pyklip.fakes.gaussfit2d(host_star_image.data, guess_index[1], guess_index[0], searchrad=7, guessfwhm=3, 
                                                        guesspeak=np.nanmax(host_star_image.data), refinefit=True)
    # host_star_counts, _ = measure_counts(ref_star_dataset[0], phot_method, None, **photometry_kwargs)
    
    # correct host star counts for  for ND if ND cal is provided
    if nd_cal is not None:
        od_val = nd_cal.interpolate_od(x_host, y_host)
        host_star_peakflux /= od_val
    

    # measure peak flux in CT dataset and use as CT=1
    _, _, ct = ct_cal.ct_excam
    max_index = np.argmax(ct)
    ct_max_frame = ct_cal.data[int(max_index)]
    guess_index = np.unravel_index(np.nanargmax(ct_max_frame), ct_max_frame.shape)
    ct_psf_peakflux, ct_psf_fwhm, _, _ = pyklip.fakes.gaussfit2d(ct_max_frame, guess_index[1], guess_index[0], searchrad=7, guessfwhm=3, 
                                                        guesspeak=np.nanmax(ct_max_frame), refinefit=True)
    # reference_psf_counts, _ = measure_counts(ref_psf_min_mask_effect, phot_method, None, **photometry_kwargs)
    
    # flux ratio between host star and CT star to scale CT PSFs
    host_to_ct_psf_ratio = (host_star_peakflux * host_star_fwhm**2) / (ct_psf_peakflux * ct_psf_fwhm**2)

    # Get the star and mask location from the PSF-subtracted image header.
    x_star = psf_sub_image.ext_hdr.get('STARLOCX')
    y_star = psf_sub_image.ext_hdr.get('STARLOCY')


    # determine x/y location of candidates
    if cand_locs is None:
        # Extract companion positions from the PSF-subtracted image and coronagraphic image headers.
        companions = parse_companions(psf_sub_image.ext_hdr)
    else:
        companions = []
        for i, cand in enumerate(cand_locs):
            x_comp = x_star + cand[0] * np.cos(np.radians(cand[1] + 90))
            y_comp = y_star + cand[0] * np.sin(np.radians(cand[1] + 90)) 
            companions.append({"id": i, "y": y_comp, "x": x_comp})

    
    # Create a for the single psf subtraction for API purposes
    psf_sub_dataset = Dataset([psf_sub_image])

    # Measure the flux of each companion
    results = []
    i = 0
    for comp_sub in companions:
        x_psf, y_psf = comp_sub['x'], comp_sub['y']
        x_psf_int = int(np.round(x_psf))
        y_psf_int = int(np.round(y_psf))

        # Compute separation and position angle relative to the star.
        dx, dy = x_psf - x_star, y_psf - y_star
        guesssep = np.hypot(dx, dy)
        guesspa = -np.degrees(np.arctan2(dx, dy)) % 360

        # Get the off-axis PSF at the planet location using the CT calibration file.
        interp_psfs, _, _ = ct_cal.GetPSF(x_psf - x_star, y_psf - y_star, psf_sub_dataset, fpam_fsam_cal)
        nearest_psf = interp_psfs[0]
        scaled_host_psf_at_planet_location = nearest_psf * host_to_ct_psf_ratio
        # measure flux of the star if it was at planet separation from mask (used in measuring flux ratio)
        guess_index = np.unravel_index(np.nanargmax(scaled_host_psf_at_planet_location), scaled_host_psf_at_planet_location.shape)
        host_planet_loc_peakflux, host_planet_loc_fwhm, _, _ = pyklip.fakes.gaussfit2d(scaled_host_psf_at_planet_location, 
                                                        guess_index[1], guess_index[0], searchrad=7, guessfwhm=ct_psf_fwhm, 
                                                        guesspeak=np.nanmax(scaled_host_psf_at_planet_location), refinefit=True)

        # Which approach to use for KLIP throughput correction.
        if thrp_corr == "L4":
            # use the klip throughput stored from PSF subtraction
            # pro: already calculated
            # con: maybe for planets of a different brightness, also must use same photonetry routine as was done for L4 KLIP throughput

            # measure flux at plaent position
            psf_sub_frame = psf_sub_image.data[kl_mode_idx]
            comp_peakflux, comp_fwhm, x_comp, y_comp = pyklip.fakes.gaussfit2d(psf_sub_frame, x_psf, y_psf, searchrad=7, 
                                                                               guessfwhm=host_planet_loc_fwhm, 
                                                                               guesspeak=psf_sub_frame[y_psf_int, x_psf_int], 
                                                                               refinefit=True)
            meas_sep_pix = np.sqrt((x_comp - x_star)**2 + (y_comp - y_star)**2)
            psf_sub_counts = np.pi * comp_peakflux * comp_fwhm**2 / 4. / np.log(2.) # total integrated flux for saving

            # interpolate algorithm throhgput from L4 data product 
            algo_thrp_data = psf_sub_image.hdu_list['KL_THRU'].data
            algo_thrp_seps = algo_thrp_data[0,:,0]
            algo_thrp = algo_thrp_data[1:][kl_mode_idx, :, 0]
            thrp_fwhms = algo_thrp_data[1:][kl_mode_idx, :, 1]

            this_thrp = np.interp(meas_sep_pix, algo_thrp_seps, algo_thrp)
            this_fwhm = np.interp(meas_sep_pix, algo_thrp_seps, thrp_fwhms)

            # correct for KLIP throughput and calculate flux ratio
            companion_host_ratio = (comp_peakflux / this_thrp) / host_planet_loc_peakflux 
        elif thrp_corr == "KLIP-FM":
            # KLIP-FM PSF forward moeling
            # Pro: PSF fit exactly for planet, exact for RDI
            # Con: linear approximation for ADI, not the best for bright planets ad ADI
            raise NotImplementedError("KLIP-FM not yet implemented")
            # #Forward model the off-axis image of the host star if it was at the planet location through the PSF subtraction process
            # kl_value, ct_value, modeled_image = forward_model_psf(
            #     coronagraphic_dataset, ref_star_dataset, ct_cal, scaled_star_psf,
            #     guesssep, guesspa, numbasis=numbasis, nwalkers=nwalkers, nburn=nburn,
            #     nsteps=nsteps, numthreads=numthreads, outputdir=output_dir, plot_results=plot_results,
            #     kl_mode_idx=kl_mode_idx
            # )
            # fm_counts_uncorrected, _ = measure_counts(modeled_image, phot_method, None, **photometry_kwargs)
            # # correct for algorithmic efficiency
            # model_counts = fm_counts_uncorrected / kl_value
            # if verbose == True:
            #     print("Host star if it was at companion ", i, " location forward modeled counts uncorrected: ", fm_counts_uncorrected)
            #     print("Host star if it was at companion ", i, " location forward modeled counts corrected: ", model_counts)
            #     print("Recovered companion ", i, " PSF sub efficiency: ", kl_value)
        elif thrp_corr == "None":
            # Don't do FM
            # Pro: simple, also can use photometry routine of choice
            # Con: doesn't account for KLIP throughput

            # Set default photometry keyword arguments if none are provided.
            photometry_kwargs = photometry_kwargs or get_photometry_kwargs(phot_method)

            # Measure counts in the PSF-subtracted image at the companion location.
            # select the frame from the KL cube
            psf_sub_frame = psf_sub_image.copy()
            psf_sub_frame.data = psf_sub_image.data[kl_mode_idx]
            psf_sub_frame.dq = psf_sub_image.dq[kl_mode_idx]
            psf_sub_frame.err = psf_sub_image.err[:,kl_mode_idx]
            psf_sub_counts, _ = measure_counts(psf_sub_frame, phot_method, (x_psf, y_psf), **photometry_kwargs)
            if verbose == True:
                print("Companion ", i, " coronagraphic, PSF-subtracted counts: ", psf_sub_counts)

            # modeled_image = simplified_psf_sub(scaled_star_psf, ct_cal, guesssep, psf_sub_efficiency)
            # measure the host star brightness. need to stick into an image class
            scaled_host_star_img = host_star_image.copy()
            scaled_host_star_img.data = scaled_host_psf_at_planet_location
            scaled_host_star_img.dq = np.zeros(scaled_host_psf_at_planet_location.shape)
            scaled_host_star_img.err = np.zeros((1,) + scaled_host_psf_at_planet_location.shape)
            host_star_counts, _ = measure_counts(scaled_host_star_img, phot_method, None, **photometry_kwargs)
            if verbose == True:
                print("Host star ", i, " simplified model counts corrected: ", host_star_counts)
        
            companion_host_ratio = psf_sub_counts / host_star_counts
        else:
            raise ValueError("{0} is not a valid option for thrp_corr".format(thrp_corr))

        # Calculate the apparent magnitude based on the host star magnitude or flux calibration.
        if host_star_in_calspec:
            apmag_data = l4_to_tda.determine_app_mag(host_star_image, host_star_image.pri_hdr['TARGET'])
            host_star_apmag = apmag_data[0].ext_hdr['APP_MAG']
            companion_mag = host_star_apmag - 2.5 * np.log10(companion_host_ratio)
        else:
            if fluxcal_factor is None:
                raise ValueError("Provide fluxcal_factor or set host_star_in_calspec=True.")
            zero_point = fluxcal_factor.ext_hdr.get('ZP')
            companion_mag = -2.5 * np.log10(psf_sub_counts) + zero_point

        # Append the results for this companion.
        results.append((
            comp_sub['id'], x_psf, y_psf, psf_sub_counts, companion_host_ratio, companion_mag
        ))
        i+=1

    # Return the results as an Astropy Table.
    return Table(
        rows=results,
        names=[
            'id', 'x', 'y',
            'measured companion counts',
            'counts_ratio',
            'companion estimated mag'
        ]
    )


def parse_companions(ext_hdr):
    """
    Parse companion positions from an image extension header.
    
    Args:
        ext_hdr (dict): Dictionary representing the image extension header.
    
    Returns:
        companions (list): List of dictionaries with keys 'id', 'x', and 'y' for each companion.
    """
    companions = []
    # Iterate over header keys to find those starting with 'SNYX'.
    for key, val in ext_hdr.items():
        if key.startswith('SNYX'):
            parts = val.split(',')
            # Ensure there are at least three parts: id, x, and y.
            if len(parts) >= 3:
                companions.append({"id": key, "y": float(parts[1]), "x": float(parts[2])})
    return companions



def measure_counts(image, phot_method, initial_xy, **kwargs):
    """
    Measure the flux counts in an image using a specified photometry method.
    
    Args:
        image (corgidrp.data.Image): Input image for photometry.
        phot_method (str): Photometry method to use ('aperture' or 'gauss2d').
        initial_xy (tuple or None): Initial (x, y) guess for centroiding.
        kwargs (dict): Arbitrary keyword arguments passed directly to the photometry method
            (e.g., fluxcal.phot_by_gauss2d_fit or fluxcal.aper_phot).
    
    Returns:
        flux (float): Measured flux in the image.
        flux_err (float): Estimated error in the measured flux.
    """
    if phot_method == 'gauss2d':
        flux, flux_err, *_ = fluxcal.phot_by_gauss2d_fit(image, centering_initial_guess=initial_xy, **kwargs)
    elif phot_method == 'aperture':
        flux, flux_err, *_ = fluxcal.aper_phot(image, centering_initial_guess=initial_xy, **kwargs)
    else:
        raise ValueError(f"Invalid photometry method: {phot_method}")
    return flux, flux_err


def get_photometry_kwargs(phot_method):
    """
    Retrieve default photometry keyword arguments based on the specified method.
    
    Args:
        phot_method (str): Photometry method to use ('aperture' or 'gauss2d').
    
    Returns:
        options[phot_method] (dict): Dictionary of default photometry keyword arguments.
    """
    common = {'centering_method': 'xy', 'centroid_roi_radius': 5}
    options = {
        'gauss2d': {'fwhm': 4, 'background_sub': True, 'r_in': 5, 'r_out': 10, **common},
        'aperture': {'encircled_radius': 7, 'frac_enc_energy': 1.0, 'subpixels': 10,
                     'background_sub': True, 'r_in': 5, 'r_out': 10, **common}
    }
    if phot_method not in options:
        raise ValueError(f"Invalid photometry method '{phot_method}'.")
    return options[phot_method]


def simplified_psf_sub(psf_frame, ct_cal, guesssep, efficiency):
    """
    TO DO: add in core throughput efficiencies?
    Perform a simplified PSF subtraction by scaling the PSF frame using a provided efficiency.
    
    Args:
        psf_frame (corgidrp.data.Image): Input PSF image.
        ct_cal (corgidrp.data.CoreThroughputCalibration): Core throughput calibration data.
        guesssep (float): Estimated separation used to determine throughput.
        efficiency (float): PSF subtraction efficiency factor.
    
    Returns:
        (corgidrp.data.Image): Scaled PSF-subtracted image.
        None (NoneType): Placeholder value.
    """
    sub_frame = psf_frame.data * efficiency
    return Image(sub_frame, pri_hdr=psf_frame.pri_hdr, ext_hdr=psf_frame.ext_hdr, err=getattr(psf_frame, 'err', None))



def forward_model_psf(
    coronagraphic_dataset,
    reference_star_dataset,
    ct_calibration,
    scaled_star_psf,
    guesssep,
    guesspa,
    outputdir=".",
    fileprefix="companion_fm",
    numbasis=[1, 2],
    mode="ADI",
    annuli=1,
    subsections=1,
    movement=1,
    inject_norm="sum",
    method="mcmc",
    stamp_size=30,
    fwhm_guess=3.0,
    nwalkers=10,
    nburn=5,
    nsteps=20,
    numthreads=1,
    star_center_err=0.05,
    platescale=21.8,
    platescale_err=0.02,
    pa_offset=0.0,
    pa_uncertainty=0.1,
    plot_results=False,
    kl_mode_idx=-1  # default to the last KL mode
):
    """
    Forward model the PSF for a companion by injecting a normalized PSF into each frame and performing KLIP subtraction.
    
    TODO: this is not fully implemented yet.

    Args:
        coronagraphic_dataset (corgidrp.data.Dataset): Dataset containing coronagraphic images.
        reference_star_dataset (corgidrp.data.Dataset): Dataset containing reference star images.
        ct_calibration (corgidrp.data.CoreThroughputCalibration): Core throughput calibration data.
        scaled_star_psf (corgidrp.data.Image): PSF subtracted image at the location of the companion,
            scaled to the host star's brightness. 
        guesssep (float): Estimated separation of the companion.
        guesspa (float): Estimated position angle (in degrees) of the companion.
        outputdir (str): Directory path to save output files.
        fileprefix (str): File prefix for output files.
        numbasis (list): List of KLIP modes to retain.
        mode (str): PSF subtraction mode (e.g., "ADI").
        annuli (int): Number of annuli to use for subtraction.
        subsections (int): Number of subsections for subtraction.
        movement (int): Movement parameter for subtraction.
        inject_norm (str): Normalization method for PSF injection.
        method (str): Fitting method (e.g., "mcmc").
        stamp_size (int): Size of the stamp used in forward modeling.
        fwhm_guess (float): Initial guess for the FWHM of the PSF.
        nwalkers (int): Number of MCMC walkers.
        nburn (int): Number of burn-in steps for MCMC.
        nsteps (int): Number of MCMC steps.
        numthreads (int): Number of threads for computation.
        star_center_err (float): Error in the star center location.
        platescale (float): Plate scale (e.g., mas/pixel).
        platescale_err (float): Error in the plate scale.
        pa_offset (float): Position angle offset.
        pa_uncertainty (float): Uncertainty in the position angle.
        plot_results (bool): Flag to enable plotting of forward modeling results.
        kl_mode_idx (int): Index of the KL mode to use (must match the one used for PSF subtraction).
    
    Returns:
        kl_value (float): KLIP throughput value extracted from the subtraction.
        ct_value (float): Core throughput value extracted from the subtraction.
        klip_image (corgidrp.data.Image): Final PSF-subtracted image after forward modeling.
    """
    amp = np.nanmax(scaled_star_psf.data)

    if plot_results == True:
        plot_dataset(coronagraphic_dataset, 'Coronagraph dataset', cmap='plasma')

    fm_dataset = coronagraphic_dataset.copy()

    # Inject the normalized PSF into each frame.
    for idx, frame in enumerate(fm_dataset):
        # TO DO: look into why this only works if I do negative guesspa
        injected_frame, _, _ = klip_fm.inject_psf(frame, ct_calibration, amp, 
                                                  guesssep, guesspa)
        fm_dataset[idx].data = injected_frame.data
    
    if plot_results == True:
        plot_dataset(fm_dataset, 'Injected PSF (fm_dataset)', cmap='plasma')

    # Perform PSF subtraction using the l3_to_l4 pipeline.
    fm_psfsub = l3_to_l4.do_psf_subtraction(
        input_dataset=fm_dataset,
        ct_calibration=ct_calibration,
        reference_star_dataset=reference_star_dataset,
        mode=mode,
        annuli=annuli,
        subsections=subsections,
        movement=movement,
        numbasis=numbasis,
        outdir=outputdir,
        fileprefix=fileprefix,
        do_crop=True,
        crop_sizexy=(100, 100),
        measure_klip_thrupt=True,
        measure_1d_core_thrupt=True
    )
    # Get the final frame from the subtraction.
    klip_data = fm_psfsub[0].data[kl_mode_idx]
    klip_thru_hdu = fm_psfsub[0].hdu_list['KL_THRU']
    klip_thru_data = klip_thru_hdu.data
    print("KL throughput", klip_thru_data)
    # Find the index of the closest separation value
    closest_idx = np.abs(klip_thru_data[0] - guesssep).argmin()

    # Get the corresponding algorithm throughput, returned data is
    # (N, n_seps, 2) where:
    # N is 1 plus the number of KL mode truncation choices (the extra “1” is for the reference separations).
    # n_seps is the number of separations sampled.
    # The last dimension (of size 2) holds the pair: (throughput, output FWHM) for each separation.
    separations = klip_thru_data[0, :, 0]  # extract separation values from the header row
    # Find the index of the separation closest to desired_sep:
    closest_idx = np.argmin(np.abs(separations - guesssep))
    kl_throughput_value = klip_thru_data[kl_mode_idx, closest_idx, 0]

    # TO DO: figure out what to do about core throughput, if anything
    ct_value = 1                    # Placeholder value for core throughput.
    ct_data = fm_psfsub[0].data[kl_mode_idx]
    ct_hdu = fm_psfsub[0].hdu_list['CT_THRU']
    ct_data = ct_hdu.data
    print("CT throughput", ct_data)
    klip_image = Image(klip_data, pri_hdr=fm_psfsub[0].pri_hdr, ext_hdr=fm_psfsub[0].ext_hdr)

    if plot_results == True:
        plot_dataset(klip_image, 'PSF-Subtracted (fm_psfsub)', cmap='plasma')

    #TO DO: don't hardcode this, ideally you can use masklocx and y
    comp_keywords = [key for key in fm_psfsub[0].ext_hdr if key.startswith("SNYX")]
    for key in comp_keywords:
        klip_image = update_companion_location_in_cropped_image(klip_image, key, (512, 512), (50, 50))

    return kl_throughput_value, ct_value, klip_image



def update_companion_location_in_cropped_image(image, comp_keyword, old_host, new_host):
    """
    Update the companion location in an image header after cropping, based on new host coordinates.
    
    Args:
        image (corgidrp.data.Image): Image containing companion location in its header.
        comp_keyword (str): Header keyword that holds the companion location.
        old_host (tuple): Original (x, y) coordinates of the host star.
        new_host (tuple): New (x, y) coordinates of the host star in the cropped image.
    
    Returns:
        image (corgidrp.data.Image): Image with updated companion location in the header.
    """
    ext_hdr = image.ext_hdr
    if comp_keyword not in ext_hdr:
        raise KeyError(f"Keyword {comp_keyword} not found in image header.")
    parts = ext_hdr[comp_keyword].split(',')
    if len(parts) < 3:
        raise ValueError(f"Unexpected format for companion location in {comp_keyword}: {ext_hdr[comp_keyword]}")
    # Since header format is "SNR,y,x", parse accordingly.
    sn_val = float(parts[0])
    old_comp_y = float(parts[1])
    old_comp_x = float(parts[2])
    # Update: host coordinates are (x,y) so use old_host[0] for x and old_host[1] for y.
    new_comp_y = int(round(old_comp_y - old_host[1] + new_host[1]))
    new_comp_x = int(round(old_comp_x - old_host[0] + new_host[0]))
    ext_hdr[comp_keyword] = f"{sn_val:5.1f},{new_comp_y:4d},{new_comp_x:4d}"
    return image



def extract_single_frame(image, frame_index=0):
    """
    Extract a single frame from a multi-frame image, including associated error and DQ arrays.
    
    Args:
        image (corgidrp.data.Image): Multi-frame image object.
        frame_index (int): Index of the frame to extract.
    
    Returns:
        new_image (corgidrp.data.Image): Extracted single frame image with error and DQ information.
    """
    data_frame = image.data[frame_index, :, :]
    # Handle error array extraction based on its dimensions.
    if image.err is not None:
        err_frame = image.err[0, frame_index, :, :] if image.err.ndim == 4 else image.err[frame_index, :, :]
    else:
        err_frame = None
    dq_frame = image.dq[frame_index, :, :] if image.dq is not None else None
    new_image = Image(data_frame, pri_hdr=image.pri_hdr, ext_hdr=image.ext_hdr, err=err_frame, dq=dq_frame)
    # Preserve specific header units if present.
    if hasattr(image, 'hdu_list'):
        new_image.hdu_list = {key: image.hdu_list[key] for key in image.hdu_list if key in ['KL_THRU', 'CT_THRU']}
    return new_image


def plot_dataset(dataset, title_prefix, cmap='plasma'):
    """
    Plots frames from a dataset in a grid layout using ZScale normalization.
    
    Function accepts either:
      - A Dataset object with a 'frames' attribute,
      - A list or tuple of frame objects (each having a 'data' attribute), or
      - A single frame object.
    
    Args:
        dataset (Dataset, list, tuple, or object): The input dataset or frame(s).
        title_prefix (str): Title prefix for each subplot.
        cmap (str): Colormap to use for the image display.
    """
    # If dataset is a Dataset object (assumed to have a 'frames' attribute), extract the frames.
    if hasattr(dataset, 'frames'):
        frames = dataset.frames
    # If dataset is not a list/tuple (i.e. a single frame), wrap it in a list.
    elif not isinstance(dataset, (list, tuple)):
        frames = [dataset]
    else:
        frames = dataset

    num_frames = len(frames)
    ncols = int(np.ceil(np.sqrt(num_frames)))
    nrows = int(np.ceil(num_frames / ncols))
    
    fig, axs = plt.subplots(nrows, ncols, figsize=(4 * ncols, 4 * nrows))
    axs = np.array(axs).flatten()
    
    for i, frame in enumerate(frames):
        data = frame.data
        data_max = data.max()
        # Use ZScaleInterval to compute the limits.
        zscale = ZScaleInterval()
        vmin, vmax = zscale.get_limits(data)
        
        im = axs[i].imshow(data, origin='lower', cmap=cmap, vmin=vmin, vmax=vmax)
        axs[i].set_title(f'{title_prefix} Frame {i} (max: {data_max:.2e})')
        plt.colorbar(im, ax=axs[i])
    
    # Turn off any unused subplots.
    for j in range(i + 1, len(axs)):
        axs[j].axis('off')
    
    fig.suptitle(f'{title_prefix} Frames', fontsize=16)
    plt.tight_layout()
    plt.show()


import pytest
import numpy as np
import corgidrp.mocks as mocks
from corgidrp.l2a_to_l2b import frame_select

np.random.seed(123)

def test_no_selection():
    """
    Tests defautl case where no frame selection is done
    """
    # use mock darks for now, doesn't really matter right now
    default_dataset = mocks.create_dark_calib_files(numfiles=5)

    pruned_dataset = frame_select(default_dataset)

    assert len(pruned_dataset) == len(default_dataset)
    assert len(pruned_dataset) == 5
    
    # assert headers are filled out
    assert pruned_dataset[0].ext_hdr['FRMSEL01'] == 1
    assert pruned_dataset[0].ext_hdr['FRMSEL02'] == False
    assert pruned_dataset[0].ext_hdr['FRMSEL03'] == None
    assert pruned_dataset[0].ext_hdr['FRMSEL04'] == None
    assert pruned_dataset[0].ext_hdr['FRMSEL05'] == None
    assert pruned_dataset[0].ext_hdr['FRMSEL06'] == None

def test_bpfrac_cutoff():
    """
    Tests the bad pixel fraction cutoff
    Cutoff is a > comparison, not >=
    """
    # use mock darks for now, doesn't really matter right now
    default_dataset = mocks.create_dark_calib_files(numfiles=5)
    default_dataset[0].dq[0:int(default_dataset.all_data.shape[-2]/2)] += 1

    # at exactly 50%, it should not drop a frame. behavior is >
    pruned_dataset = frame_select(default_dataset, bpix_frac=0.5)
    assert len(pruned_dataset) == len(default_dataset)
    assert len(pruned_dataset) == 5

    # one pixel over 50%, we should drop the first frame
    default_dataset[0].dq[-1,-1] += 1
    pruned_dataset = frame_select(default_dataset, bpix_frac=0.5)
    assert len(pruned_dataset) != len(default_dataset)
    assert len(pruned_dataset) == 4
    assert default_dataset[0].filename in pruned_dataset[0].ext_hdr['HISTORY'][-1] # test history
    assert pruned_dataset[0].ext_hdr['FRMSEL01'] == 0.5

    # allowing DQ = 1 should make no frames get dropped
    pruned_dataset = frame_select(default_dataset, bpix_frac=0.5, allowed_bpix=1)
    assert len(pruned_dataset) == len(default_dataset)
    assert len(pruned_dataset) == 5

    # allowing DQ = 1 should not affect DQ = 2
    default_dataset[0].dq[0:int(default_dataset.all_data.shape[-2]/2)] = 2
    default_dataset[0].dq[-1,-1] = 2
    pruned_dataset = frame_select(default_dataset, bpix_frac=0.5, allowed_bpix=1)
    assert len(pruned_dataset) != len(default_dataset)
    assert len(pruned_dataset) == 4
    assert default_dataset[0].filename in pruned_dataset[0].ext_hdr['HISTORY'][-1] # test history


def test_overexp():
    """
    Tests the overexp keyword
    """
    default_dataset = mocks.create_dark_calib_files(numfiles=5)
    default_dataset[0].ext_hdr['OVEREXP'] = True

    # doest nothing
    pruned_dataset = frame_select(default_dataset)
    assert len(pruned_dataset) == len(default_dataset)
    assert len(pruned_dataset) == 5

    # remove overexp
    pruned_dataset = frame_select(default_dataset, overexp=True)
    assert len(pruned_dataset) != len(default_dataset)
    assert len(pruned_dataset) == 4
    assert pruned_dataset[0].ext_hdr['FRMSEL02'] == True

def test_tt_rms():
    """
    Tests for tt rms jitter
    """
    default_dataset = mocks.create_dark_calib_files(numfiles=5)
    # add tt rms header
    tt_rms = 0
    for frame in default_dataset:
        frame.ext_hdr['Z2VAR'] = tt_rms
        frame.ext_hdr['Z3VAR'] = tt_rms
        tt_rms += 1

    # does nothing
    pruned_dataset = frame_select(default_dataset)
    assert len(pruned_dataset) == len(default_dataset)
    assert len(pruned_dataset) == 5

    # removes 2 frames
    pruned_dataset = frame_select(default_dataset, tt_rms_thres=2.5)
    assert len(pruned_dataset) != len(default_dataset)
    assert len(pruned_dataset) == 3
    assert pruned_dataset[0].ext_hdr['FRMSEL03'] == 2.5
    assert pruned_dataset[0].ext_hdr['FRMSEL04'] == 2.5

def test_tt_bias():
    """
    Tests for tt offset
    """
    default_dataset = mocks.create_dark_calib_files(numfiles=5)
    # add tt rms header
    tt_rms = 0
    for frame in default_dataset:
        frame.ext_hdr['Z2RES'] = tt_rms
        frame.ext_hdr['Z3RES'] = tt_rms
        tt_rms += 1

    # does nothing
    pruned_dataset = frame_select(default_dataset)
    assert len(pruned_dataset) == len(default_dataset)
    assert len(pruned_dataset) == 5

    # removes 2 frames
    pruned_dataset = frame_select(default_dataset, tt_bias_thres=2.5)
    assert len(pruned_dataset) != len(default_dataset)
    assert len(pruned_dataset) == 3
    assert pruned_dataset[0].ext_hdr['FRMSEL05'] == 2.5
    assert pruned_dataset[0].ext_hdr['FRMSEL06'] == 2.5

def test_remove_all():
    """
    Uses multiple methods to remove frames.
    Tests that error is raised if all frames is removed
    """
    # use mock darks for now, doesn't really matter right now
    default_dataset = mocks.create_dark_calib_files(numfiles=5)
    # make the first frame have 50% bad pixels
    default_dataset[0].dq[0:int(default_dataset.all_data.shape[-2]/2)] += 1
    # add tt rms header
    tt_rms = 0
    for frame in default_dataset:
        frame.ext_hdr['Z2VAR'] = tt_rms
        frame.ext_hdr['Z3VAR'] = tt_rms
        tt_rms += 1
    # add overexp
    default_dataset[1].ext_hdr['OVEREXP'] = True

    # keep only 1 frame
    # bpix_frac removes index 0
    # overexp removes index 1
    # tt_rms_thres removes indicies 3,4
    pruned_dataset = frame_select(default_dataset, bpix_frac=0.1, overexp=True, tt_rms_thres=2.5)
    assert len(pruned_dataset) == 1

    # removes all frames
    with pytest.raises(ValueError):
        pruned_dataset = frame_select(default_dataset, bpix_frac=0.1, overexp=True, tt_rms_thres=1.5)

def test_marking():
    """
    Tests marking frames as bad instead of completely dropping them
    """
    default_dataset = mocks.create_dark_calib_files(numfiles=5)
    # add tt rms header
    tt_rms = 0
    for frame in default_dataset:
        frame.ext_hdr['Z2RES'] = tt_rms
        frame.ext_hdr['Z3RES'] = tt_rms
        tt_rms += 1

    # does nothing
    pruned_dataset = frame_select(default_dataset, discard_bad=False)
    assert len(pruned_dataset) == len(default_dataset)
    assert len(pruned_dataset) == 5
    for frame in pruned_dataset:
        assert frame.ext_hdr['IS_BAD'] == False

    # marks 2 frames as bad
    pruned_dataset = frame_select(default_dataset, tt_bias_thres=2.5, discard_bad=False)
    assert len(pruned_dataset) == len(default_dataset)
    # but first 3 are good
    for frame in pruned_dataset[:3]:
        assert frame.ext_hdr['IS_BAD'] == False
    # and last 2 are bad
    for frame in pruned_dataset[3:]:
        assert frame.ext_hdr['IS_BAD'] == True

if __name__ == "__main__":
    test_no_selection()
    test_bpfrac_cutoff()
    test_overexp()
    test_tt_rms()
    test_remove_all()


import numpy as np
import pytest
from corgidrp.data import Dataset, Image
from corgidrp.l3_to_l4 import crop
from corgidrp.mocks import create_default_L3_headers

def make_test_dataset(shape=[100,100],centxy=None):
    """
    Make 2D or 3D test data.

    Args:
        shape (arraylike, optional): data shape. Defaults to [100,100].
        centxy (arraylike,optional): location of 4 pixel dot. Defaults to center of array.

    Returns:
        corgidrp.data.Dataset: test data with a 2x2 "PSF" at location centxy.
    """
    shape = np.array(shape)

    test_arr = np.zeros(shape)
    if centxy is None:
        cent = np.array(shape)/2 - 0.5
    else:
        cent = [centxy[-i] for i in np.array(range(len(centxy)))+1]
        
    prihdr,exthdr = create_default_L3_headers()
    exthdr['STARLOCX'] = cent[1]
    exthdr['STARLOCY'] = cent[0]
    exthdr['LSAM_H'] = cent[1]
    exthdr['LSAM_V'] = cent[0]
    exthdr['LSAMNAME'] = 'NFOV'
    
    if len(shape) == 2:
        test_arr[int(cent[0]-0.5):int(cent[0]+1.5),int(cent[1]-0.5):int(cent[1]+1.5)] = 1

    elif len(shape) == 3:
        test_arr[:,int(cent[0]-0.5):int(cent[0]+1.5),int(cent[1]-0.5):int(cent[1]+1.5)] = 1

    test_dataset = Dataset([Image(test_arr,prihdr,exthdr)])

    return test_dataset

goal_arr = np.zeros((10,10))
goal_arr[4:6,4:6] = 1

goal_rect_arr = np.zeros((10,20))
goal_rect_arr[4:6,9:11] = 1

def test_2d_square_center_crop():
    """ Test cropping to the center of a square using the header keywords "STARLOCX/Y".
    """

    test_dataset = make_test_dataset(shape=[100,100],centxy=[49.5,49.5])
    cropped_test_dataset = crop(test_dataset,sizexy=10,centerxy=None)

    if not cropped_test_dataset[0].data == pytest.approx(goal_arr):
        raise Exception("Unexpected result for 2D square crop test.")

def test_manual_center_crop():
    """ Test overriding crop location using centerxy argument and make sure 
    DETPIX0X/Y header keyword is updated correctly.
    """

    test_dataset = make_test_dataset(shape=[12,12],centxy=[5.5,5.5])
    cropped_test_dataset = crop(test_dataset,sizexy=10,centerxy=[6.5,6.5])

    offset_goal_arr = np.zeros((10,10))
    offset_goal_arr[3:5,3:5] = 1

    expected_detpix_xy = (2,2)

    if not (cropped_test_dataset[0].ext_hdr["DETPIX0X"],
            cropped_test_dataset[0].ext_hdr["DETPIX0Y"]) == expected_detpix_xy:
        raise Exception("Extension header DETPIX0X/Y not updated correctly.")
    

    if not cropped_test_dataset[0].data == pytest.approx(offset_goal_arr):
        raise Exception("Unexpected result for manual crop test.")

def test_2d_square_offcenter_crop():
    """ Test cropping off-center square data.
    """

    test_dataset = make_test_dataset(shape=[100,100],centxy=[24.5,49.5])
    cropped_test_dataset = crop(test_dataset,sizexy=10,centerxy=None)

    if not cropped_test_dataset[0].data == pytest.approx(goal_arr):
        raise Exception("Unexpected result for 2D square offcenter crop test.")

def test_2d_rect_offcenter_crop():
    """ Tests cropping off-center non-square data.
    """
    test_dataset = make_test_dataset(shape=[100,40],centxy=[24.5,49.5])
    cropped_test_dataset = crop(test_dataset,sizexy=[20,10],centerxy=None)

    if not cropped_test_dataset[0].data == pytest.approx(goal_rect_arr):
        raise Exception("Unexpected result for 2D rect offcenter crop test.")

def test_3d_rect_offcenter_crop():
    """ Tests cropping 3D off-center non-square data.
    """
    test_dataset = make_test_dataset(shape=[3,100,40],centxy=[24.5,49.5])
    cropped_test_dataset = crop(test_dataset,sizexy=[20,10],centerxy=None)

    goal_rect_arr3d = np.array([goal_rect_arr,goal_rect_arr,goal_rect_arr])

    if not cropped_test_dataset[0].data == pytest.approx(goal_rect_arr3d):
        raise Exception("Unexpected result for 2D rect offcenter crop test.")

def test_edge_of_detector():
    """ Tests that trying to crop a region right at the edge of the 
    detector succeeds.
    """
    test_dataset = make_test_dataset(shape=[100,100],centxy=[94.5,94.5])
    cropped_test_dataset = crop(test_dataset,sizexy=10,centerxy=None)

    if not cropped_test_dataset[0].data == pytest.approx(goal_arr):
        raise Exception("Unexpected result for edge of FOV crop test.")

def test_outside_detector_edge():
    """ Tests that trying to crop a region outside the detector fails.
    """

    test_dataset = make_test_dataset(shape=[100,100],centxy=[95.5,95.5])

    with pytest.raises(ValueError):
        _ = crop(test_dataset,sizexy=10,centerxy=None)

def test_nonhalfinteger_centxy():
    """ Tests that trying to crop data to a center that is not at the intersection 
    of 4 pixels results in centering on the nearest pixel intersection.
    """
    test_dataset = make_test_dataset(shape=[100,100],centxy=[49.5,49.5])
    cropped_test_dataset = crop(test_dataset,sizexy=10,centerxy=[49.7,49.7])

    if not cropped_test_dataset[0].data == pytest.approx(goal_arr):
        raise Exception("Unexpected result for non half-integer crop test.")

def test_header_updates_2d():
    """ Tests that cropping works, and updates header values related to 
    pixel positions and data shapes correctly, for 3D data.
    """
    
    test_dataset = make_test_dataset(shape=[100,100],centxy=[49.5,49.5])
    test_dataset[0].ext_hdr["STARLOCX"] = 49.5
    test_dataset[0].ext_hdr["STARLOCY"] = 49.5
    test_dataset[0].pri_hdr["CRPIX1"] = 50.5
    test_dataset[0].pri_hdr["CRPIX2"] = 50.5
    
    cropped_test_dataset = crop(test_dataset,sizexy=10,centerxy=None)

    if not cropped_test_dataset[0].ext_hdr["STARLOCX"] == 4.5:
        raise Exception("Frame header kw STARLOCX not updated correctly.")
    if not cropped_test_dataset[0].ext_hdr["STARLOCY"] == 4.5:
        raise Exception("Frame header kw STARLOCY not updated correctly.")
    if not cropped_test_dataset[0].pri_hdr["CRPIX1"] == 5.5:
        raise Exception("Frame header kw CRPIX1 not updated correctly.")
    if not cropped_test_dataset[0].pri_hdr["CRPIX2"] == 5.5:
        raise Exception("Frame header kw CRPIX2 not updated correctly.")
    if not cropped_test_dataset[0].ext_hdr["NAXIS1"] == 10:
        raise Exception("Frame header kw NAXIS1 not updated correctly.")
    if not cropped_test_dataset[0].ext_hdr["NAXIS2"] == 10:
        raise Exception("Frame header kw NAXIS2 not updated correctly.")
    if not cropped_test_dataset[0].err_hdr["NAXIS1"] == 10:
        raise Exception("Frame err header kw NAXIS1 not updated correctly.")
    if not cropped_test_dataset[0].err_hdr["NAXIS2"] == 10:
        raise Exception("Frame err header kw NAXIS2 not updated correctly.")
    if not cropped_test_dataset[0].dq_hdr["NAXIS1"] == 10:
        raise Exception("Frame dq header kw NAXIS1 not updated correctly.")
    if not cropped_test_dataset[0].dq_hdr["NAXIS2"] == 10:
        raise Exception("Frame dq header kw NAXIS2 not updated correctly.")
    
def test_header_updates_3d():
    """ Tests that cropping works, and updates header values related to pixel 
    positions and data shapes correctly, for 3D data.
    """
    
    test_dataset = make_test_dataset(shape=[3,100,100],centxy=[49.5,49.5])
    test_dataset[0].ext_hdr["STARLOCX"] = 49.5
    test_dataset[0].ext_hdr["STARLOCY"] = 49.5
    test_dataset[0].pri_hdr["CRPIX1"] = 50.5
    test_dataset[0].pri_hdr["CRPIX2"] = 50.5
    
    cropped_test_dataset = crop(test_dataset,sizexy=10,centerxy=None)

    if not cropped_test_dataset[0].ext_hdr["STARLOCX"] == 4.5:
        raise Exception("Frame header kw STARLOCX not updated correctly.")
    if not cropped_test_dataset[0].ext_hdr["STARLOCY"] == 4.5:
        raise Exception("Frame header kw STARLOCY not updated correctly.")
    if not cropped_test_dataset[0].pri_hdr["CRPIX1"] == 5.5:
        raise Exception("Frame header kw CRPIX1 not updated correctly.")
    if not cropped_test_dataset[0].pri_hdr["CRPIX2"] == 5.5:
        raise Exception("Frame header kw CRPIX2 not updated correctly.")
    if not cropped_test_dataset[0].ext_hdr["NAXIS1"] == 10:
        raise Exception("Frame header kw NAXIS1 not updated correctly.")
    if not cropped_test_dataset[0].ext_hdr["NAXIS2"] == 10:
        raise Exception("Frame header kw NAXIS2 not updated correctly.")
    if not cropped_test_dataset[0].ext_hdr["NAXIS3"] == 3:
        raise Exception("Frame header kw NAXIS3 not updated correctly.")
    if not cropped_test_dataset[0].dq_hdr["NAXIS1"] == 10:
        raise Exception("Frame dq header kw NAXIS1 not updated correctly.")
    if not cropped_test_dataset[0].dq_hdr["NAXIS2"] == 10:
        raise Exception("Frame dq header kw NAXIS2 not updated correctly.")
    if not cropped_test_dataset[0].dq_hdr["NAXIS3"] == 3:
        raise Exception("Frame dq header kw NAXIS3 not updated correctly.")
    if not cropped_test_dataset[0].err_hdr["NAXIS1"] == 10:
        raise Exception("Frame err header kw NAXIS1 not updated correctly.")
    if not cropped_test_dataset[0].err_hdr["NAXIS2"] == 10:
        raise Exception("Frame err header kw NAXIS2 not updated correctly.")
    if not cropped_test_dataset[0].err_hdr["NAXIS3"] == 3:
        raise Exception("Frame err header kw NAXIS3 not updated correctly.")
    
def test_non_nfov_input():
    """ Crop function is not configured for non-NFOV observations and should
    fail if the Lyot stop is not in the narrow FOV position, unless the desired
    image size is provided manually.
    """

    test_dataset = make_test_dataset(shape=[100,100],centxy=[50.5,50.5])
    for frame in test_dataset:
        frame.ext_hdr['LSAMNAME'] = 'WFOV'

    try:
        _ = crop(test_dataset,sizexy=20,centerxy=None)
    except:
        raise ValueError('Cropping a non-NFOV observation failed even though sizexy was provided')
    
    
    with pytest.raises(UserWarning):
        _ = crop(test_dataset,sizexy=None,centerxy=None)

def test_detpix0_nonzero():
    """ Tests that the detector pixel header keyword is updated correctly if it 
    already exists and is nonzero.
    """
    test_dataset = make_test_dataset(shape=[100,40],centxy=[24.5,49.5])
    test_dataset[0].ext_hdr.set('DETPIX0X',24)
    test_dataset[0].ext_hdr.set('DETPIX0Y',34)

    expected_detpix_xy = (39,79)
    
    cropped_test_dataset = crop(test_dataset,sizexy=[20,10],centerxy=None)

    if not cropped_test_dataset[0].data == pytest.approx(goal_rect_arr):
        raise Exception("Unexpected result for 2D rect offcenter crop test with nonzero DETPIX0X/Y.")
    
    if not (cropped_test_dataset[0].ext_hdr["DETPIX0X"],
            cropped_test_dataset[0].ext_hdr["DETPIX0Y"]) == expected_detpix_xy:
        raise Exception("Extension header DETPIX0X/Y not updated correctly.")
    

if __name__ == "__main__":
    # test_2d_square_center_crop()
    test_manual_center_crop()
    # test_2d_square_offcenter_crop()
    # test_2d_rect_offcenter_crop()
    # test_3d_rect_offcenter_crop()
    # test_edge_of_detector()
    # test_outside_detector_edge()
    # test_nonhalfinteger_centxy()
    # test_header_updates_2d()
    # test_header_updates_3d()
    # test_non_nfov_input()
    test_detpix0_nonzero()




import numpy as np
from scipy.optimize import minimize

# Define default parameters for the satellite spot finding algorithm
# Spot separation in pixels
# separation in lambda/D and multiply by pixel per lambda/D
spot_separation_nfov = 6.5*(51.46*0.575/13) # 14.79
spot_separation_spec = 6.0*(51.46*0.730/13) # 17.34
spot_separation_wfov = 13*(51.46*0.825/13) # 42.4545

satellite_spot_parameters_defaults = {
    "NFOV": {
        "offset": {
            "spotSepPix": spot_separation_nfov,
            "roiRadiusPix": 4.5,
            "probeRotVecDeg": [0, 90],
            "nSubpixels": 100,
            "nSteps": 7,
            "stepSize": 1,
            "nIter": 6,
        },
        "separation": {
            "spotSepPix": spot_separation_nfov,
            "roiRadiusPix": 1.5,
            "probeRotVecDeg": [0, 90],
            "nSubpixels": 100,
            "nSteps": 21,
            "stepSize": 0.25,
            "nIter": 5,
        }
    },
    "SPEC660": {
        "offset": {
            "spotSepPix": spot_separation_spec,
            "roiRadiusPix": 6,
            "probeRotVecDeg": [0,],
            "nSubpixels": 100,
            "nSteps": 9,
            "stepSize": 1,
            "nIter": 6,
        },
        "separation": {
            "spotSepPix": spot_separation_spec,
            "roiRadiusPix": 4,
            "probeRotVecDeg": [0,],
            "nSubpixels": 100,
            "nSteps": 21,
            "stepSize": 0.25,
            "nIter": 5,
        }
    },
    "SPEC730": {
        "offset": {
            "spotSepPix": spot_separation_spec,
            "roiRadiusPix": 6,
            "probeRotVecDeg": [0,],
            "nSubpixels": 100,
            "nSteps": 9,
            "stepSize": 1,
            "nIter": 6,
        },
        "separation": {
            "spotSepPix": spot_separation_spec,
            "roiRadiusPix": 4,
            "probeRotVecDeg": [0,],
            "nSubpixels": 100,
            "nSteps": 21,
            "stepSize": 0.25,
            "nIter": 5,
        }
    },
    "WFOV": {
        "offset": {
            "spotSepPix": spot_separation_wfov,
            "roiRadiusPix": 4.5,
            "probeRotVecDeg": [0, 90],
            "nSubpixels": 100,
            "nSteps": 7,
            "stepSize": 1,
            "nIter": 6,
        },
        "separation": {
            "spotSepPix": spot_separation_wfov,
            "roiRadiusPix": 4.5,
            "probeRotVecDeg": [0, 90],
            "nSubpixels": 100,
            "nSteps": 21,
            "stepSize": 0.25,
            "nIter": 5,
        }
    }
}


def validate_satellite_spot_parameters(params):
    """
    Checks if a dictionary conforms to the required satellite spot parameters format.

    Args:
        params (dict): Dictionary to validate.

    Returns:
        bool: True if valid, False otherwise.
    """
    required_structure = {
        "offset": {
            "spotSepPix": float,
            "roiRadiusPix": float,
            "probeRotVecDeg": list,
            "nSubpixels": int,
            "nSteps": int,
            "stepSize": float,
            "nIter": int,
        },
        "separation": {
            "spotSepPix": float,
            "roiRadiusPix": float,
            "probeRotVecDeg": list,
            "nSubpixels": int,
            "nSteps": int,
            "stepSize": float,
            "nIter": int,
        }
    }

    if not isinstance(params, dict):
        return False

    for section in ['offset', 'separation']:
        if section not in params or not isinstance(params[section], dict):
            return False
        for key, expected_type in validate_satellite_spot_parameters.__annotations__[section].items():
            if key not in params[section]:
                return False
            if not isinstance(params[section][key], expected_type):
                return False

    return True


def update_parameters(params, new_values):
    """
    Updates a nested dictionary of parameters with new values.

    Args:
        params (dict):
            Original nested dictionary containing initial parameter values.
            The structure should be: {key1: {subkey1: val, subkey2: val}, key2: {...}}
        new_values (dict):
            Nested dictionary containing new parameter values to update.
            Must match the structure of `params`.

    Returns:
        dict: The updated nested dictionary.

    Raises:
        KeyError: If `new_values` contains keys or subkeys not present in `params`.

    Example:
        >>> params = {'a': {'x': 1, 'y': 2}, 'b': {'z': 3}}
        >>> new_values = {'a': {'x': 10}}
        >>> update_parameters(params, new_values)
        {'a': {'x': 10, 'y': 2}, 'b': {'z': 3}}

        >>> new_values = {'b': {'w': 5}}
        >>> update_parameters(params, new_values)
        KeyError: "Subkey 'w' is not a valid parameter under key 'b'."
    """
    for key, subdict in new_values.items():
        if key not in params:
            raise KeyError(f"Key '{key}' is not a valid parameter.")

        for subkey, value in subdict.items():
            if subkey not in params[key]:
                raise KeyError(f"Subkey '{subkey}' is not a valid parameter under key '{key}'.")

            params[key][subkey] = value

    return params


def circle(nx, ny, roiRadiusPix, xShear, yShear, nSubpixels=100):
    """
    Generates a circular aperture with an antialiased edge at specified offsets.

    This function is used as a software window for isolating a region of interest. 
    Grayscale edges are applied to account for low detector sampling, ensuring that 
    fractional values along the edges are preserved.

    Args:
        nx (int):  
            Width of the 2D array to create.  
        ny (int):  
            Height of the 2D array to create.  
        roiRadiusPix (float):  
            Radius of the circle in pixels.  
        xShear (float):  
            Lateral offset of the circle's center from the array's center pixel along the x-axis.  
        yShear (float):  
            Lateral offset of the circle's center from the array's center pixel along the y-axis.  
        nSubpixels (int, optional):  
            Number of subpixels per edge pixel used for antialiasing. Each edge pixel is subdivided 
            into a square subarray of `nSubpixels` across, where binary values are assigned and 
            averaged to produce grayscale edge values between 0 and 1.  
            Defaults to 100. Must be a positive integer.  

    Returns:
        numpy.ndarray:  
            A 2D array containing the generated circular aperture mask.  
    """

    if nx % 2 == 0:
        x = np.linspace(-nx / 2.0, nx / 2.0 - 1, nx) - xShear
    elif nx % 2 == 1:
        x = np.linspace(-(nx - 1) / 2.0, (nx - 1) / 2.0, nx) - xShear

    if ny % 2 == 0:
        y = np.linspace(-ny / 2.0, ny / 2.0 - 1, ny) - yShear
    elif ny % 2 == 1:
        y = np.linspace(-(ny - 1) / 2.0, (ny - 1) / 2.0, ny) - yShear

    dx = x[1] - x[0]
    [X, Y] = np.meshgrid(x, y)
    RHO = np.sqrt(X * X + Y * Y)

    halfWindowWidth = np.sqrt(2.0) * dx
    mask = -1 * np.ones(RHO.shape)
    mask[np.abs(RHO) < roiRadiusPix - halfWindowWidth] = 1
    mask[np.abs(RHO) > roiRadiusPix + halfWindowWidth] = 0
    grayInds = np.array(np.nonzero(mask == -1))
    # print('Number of grayscale points = %d' % grayInds.shape[1])

    dxHighRes = 1.0 / float(nSubpixels)
    xUp = (
        np.linspace(-(nSubpixels - 1) / 2.0, (nSubpixels - 1) / 2.0, nSubpixels)
        * dxHighRes
    )
    [Xup, Yup] = np.meshgrid(xUp, xUp)

    subpixelArray = np.zeros((nSubpixels, nSubpixels))
    # plt.figure(); plt.imshow(RHO); plt.colorbar(); plt.pause(0.1)

    # Compute the value between 0 and 1 of each edge pixel along the circle by
    # taking the mean of the binary subpixels.
    for iInterior in range(grayInds.shape[1]):
        subpixelArray = 0 * subpixelArray

        xCenter = X[grayInds[0, iInterior], grayInds[1, iInterior]]
        yCenter = Y[grayInds[0, iInterior], grayInds[1, iInterior]]
        RHOHighRes = np.sqrt((Xup + xCenter) ** 2 + (Yup + yCenter) ** 2)
        # plt.figure(); plt.imshow(RHOHighRes); plt.colorbar(); plt.pause(1/20)

        subpixelArray[RHOHighRes <= roiRadiusPix] = 1
        pixelValue = np.sum(subpixelArray) / float(nSubpixels * nSubpixels)
        mask[grayInds[0, iInterior], grayInds[1, iInterior]] = pixelValue

    return mask


def find_optimum_1d(xVec, arrayToFit):
    """
    Fits a parabola to a 1D array and returns the location of the minimum or maximum.

    Args:
        xVec (array_like):  
            1D array of coordinate values corresponding to `arrayToFit`.  
        arrayToFit (array_like):  
            1D array of values to fit.  

    Returns:
        xOpt (float):  
            Best-fit value for the optimum (i.e., minimum or maximum) of the parabola.  
    """

    if len(xVec) != len(arrayToFit):
        raise ValueError("Both input arrays must have same length.")

    # Normalize arrayToFit to avoid spurious numerical precision issues.
    arrayToFit = arrayToFit / np.max(arrayToFit)

    orderOfFit = 2
    p0, p1, _ = np.polyfit(xVec, arrayToFit, orderOfFit)

    # Avoid divide by zero error
    if np.abs(p0) < np.finfo(p0).eps:
        raise ValueError(
            "Quadratic fit failed because data being "
            + "fitted has no quadratic component."
        )

    # Find x value at the optimum
    xOpt = -p1 / (2 * p0)

    return xOpt


def find_optimum_2d(xVec, yVec, arrayToFit, mask):
    """
    Fits a paraboloid to a 2D array and returns the location of the minimum or maximum.

    Args:
        xVec (array_like):  
            1D array of coordinate values along axis 1 of `arrayToFit`.  
        yVec (array_like):  
            1D array of coordinate values along axis 0 of `arrayToFit`.  
        arrayToFit (array_like):  
            2D array of values to fit.  
        mask (array_like):  
            2D boolean mask indicating which pixels to use. Must have the same shape as `arrayToFit`.  

    Returns:
        tuple:  
            - `xBest` (float): Best-fit value along axis 1.  
            - `yBest` (float): Best-fit value along axis 0.  

    Notes:
        - Modified from code at  
          [MATLAB Central](https://au.mathworks.com/matlabcentral/answers/5482-fit-a-3d-curve).  
        - Based on equations from  
          [Math StackExchange](https://math.stackexchange.com/questions/2010758/how-do-i-fit-a-paraboloid-surface-to-nine-points-and-find-the-minimum).  
    """

    if arrayToFit.shape != mask.shape:
        raise ValueError("arrayToFit and mask must have same shape")
    nx = len(xVec)
    ny = len(yVec)
    if arrayToFit.shape[0] != ny:
        raise ValueError("yVec and axis 0 of arrayToFit must have same length")
    if arrayToFit.shape[1] != nx:
        raise ValueError("xVec and axis 1 of arrayToFit must have same length")

    # Normalize arrayToFit to avoid spurious numerical precision issues.
    arrayToFit = arrayToFit / np.max(arrayToFit)

    maskBool = np.asarray(mask).astype(bool)
    nPix = np.sum(mask.astype(int))

    #  Set the basis functions
    [X, Y] = np.meshgrid(xVec, yVec)
    f0 = X**2
    f1 = X * Y
    f2 = Y**2
    f3 = X
    f4 = Y
    f5 = np.ones((ny, nx))

    # Write as matrix equation and solve for coefficients
    A = np.concatenate(
        (
            f0[maskBool].reshape((nPix, 1)),
            f1[maskBool].reshape((nPix, 1)),
            f2[maskBool].reshape((nPix, 1)),
            f3[maskBool].reshape((nPix, 1)),
            f4[maskBool].reshape((nPix, 1)),
            f5[maskBool].reshape((nPix, 1)),
        ),
        axis=1,
    )
    y = arrayToFit[maskBool].flatten()
    temp = np.linalg.lstsq(A, y, rcond=None)
    coeffs = temp[0]

    # Take partial derivatives w.r.t. x and y, set to zero, and solve for
    # x and y.
    # G = 0 = a*X**2 + b*X*Y + c*Y**2 + d*X + e*Y + f
    # @G/@X = 2*a*X + b*Y + d = 0 --> Solve for X. Use eq. for Y below.
    # @G/@Y = b*X + 2*c*Y + e = 0 --> Solve for Y in terms of X.
    a, b, c, d, e = coeffs[0:5]

    # Avoid divide by zero errors
    x_denominator = b * b - 4.0 * a * c
    y_denominator = 2 * c
    # The factors 4 and 2 come from the values in the denomimators
    if (
        np.abs(x_denominator) < 4 * np.finfo(x_denominator).resolution
        or np.abs(y_denominator) < 2 * np.finfo(y_denominator).resolution
    ):
        raise ValueError(
            "Quadratic fit failed because data being "
            + "fitted has no quadratic component."
        )

    xBest = (2.0 * c * d - e * b) / (b * b - 4.0 * a * c)
    yBest = -(b * xBest + e) / (2 * c)

    return xBest, yBest


def _roi_mask_for_spots(optim_params, fixed_params):
    """
    Computes the stellar offset and spot separation simultaneously.

    Args:
        optim_params (list):  
            List of parameters to be optimized by `scipy.optimize.minimize()`.  
            These three variables must be provided in this exact order:  
            `[xOffsetGuess, yOffsetGuess, spotSepGuessPix]`          
            - `xOffsetGuess` (float): Initial guess for the star offset from the array center pixel in the x direction. Units: pixels.  
            - `yOffsetGuess` (float): Initial guess for the star offset from the array center pixel in the y direction. Units: pixels.  
            - `spotSepGuessPix` (float): Initial guess for the spot separation from the star center. Units: pixels.  
        fixed_params (list):  
            List of fixed parameters used in computing the cost function.  
            These four variables must be provided in this exact order:  
            `[spotArray, probeRotVecDeg, roiRadiusPix, nSubpixels]`         
            - `spotArray` (numpy.ndarray):  
              2D array of DM-generated satellite spots.  
              Computed externally from probed images as `(Iplus + Iminus)/2 - Iunprobed`.  
            - `probeRotVecDeg` (array_like):  
              1D array of angles (in degrees) counterclockwise from the x-axis  
              to rotate the regions of interest when summing satellite spots.  
              A pair of satellite spots is represented by a single value.  
              - Example for a single pair along the x-axis: `[0]` (not `[0, 180]`).  
              - Example for a plus-shaped layout of spots: `[0, 90]`.  
            - `roiRadiusPix` (float):  
              Radius of each region of interest used when summing the intensity  
              of a satellite spot. Units: pixels.  
            - `nSubpixels` (int):  
              Number of subpixels used to refine edge values of the region-of-interest mask.  
              The value of the edge pixels in the ROI is the mean of all subpixel values.  

    Returns:
        tuple:  
            - `cost` (float):  
              The summed total intensity passing through the region-of-interest mask,  
              multiplied by `-1`. Used as the cost function to be minimized by `scipy.optimize.minimize()`.  
            - `roi_mask` (numpy.ndarray):  
              2D array with values between 0 and 1 (inclusive), indicating how much  
              to weight the intensity in each pixel in the summed cost.  
    """

    # Unpack and check the input list of optimization variables
    try:
        xOffsetGuess, yOffsetGuess, spotSepGuessPix = optim_params
    except:
        raise ValueError("optim_params must be an iterable of length 3")

    # Unpack and check the values in the list of fixed inputs.
    try:
        spotArray, probeRotVecDeg, roiRadiusPix, nSubpixels = fixed_params
    except:
        raise ValueError("fixed_params must be an iterable of length 4")

    # Generate mask to isolate the satellite spots in spotArray
    ny, nx = spotArray.shape
    roi_mask = np.zeros((ny, nx))
    for iRot, rotDeg in enumerate(probeRotVecDeg):
        rotRad = np.radians(rotDeg)
        xProbePlusCoord = np.array(
            [
                np.sin(rotRad) * spotSepGuessPix + yOffsetGuess,
                np.cos(rotRad) * spotSepGuessPix + xOffsetGuess,
            ]
        )
        rotRad += np.pi
        xProbeMinusCoord = np.array(
            [
                np.sin(rotRad) * spotSepGuessPix + yOffsetGuess,
                np.cos(rotRad) * spotSepGuessPix + xOffsetGuess,
            ]
        )

        roi_mask += circle(
            nx,
            ny,
            roiRadiusPix,
            xProbePlusCoord[1],
            xProbePlusCoord[0],
            nSubpixels=nSubpixels,
        )
        roi_mask += circle(
            nx,
            ny,
            roiRadiusPix,
            xProbeMinusCoord[1],
            xProbeMinusCoord[0],
            nSubpixels=nSubpixels,
        )

    # Cost is negative because scipy minimizes the cost function
    # but we want to maximize the sum.
    cost = -np.sum(roi_mask * spotArray)

    return cost, roi_mask


def _cost_func_spots(optim_params, fixed_params):
    """
    Return only the cost value for scipy.optimize.minimize().
        
    This is a thin wrapper for occastro.roi_mask_for_spots() because
    scipy.optimize.minimize() requires the given function to return only
    the cost function value but we want some other diagnostic outputs from
    occastro.roi_mask_for_spots() as well.

    Args:
        optim_params (tuple): Tuple of optimization parameters.
        fixed_params (dict): Dictionary of fixed parameters.

    Returns:
        float: The cost function value.
    """

    cost, _ = _roi_mask_for_spots(optim_params, fixed_params)

    return cost


def calc_star_location_and_spot_separation(
    spotArray, xOffsetGuess, yOffsetGuess, tuningParamDict
):
    """
    Calculates the center of the occulted star using satellite spots.

    A single processed image of satellite spots is used. Multiple software masks 
    are generated and applied to the measured spots to determine the stellar location. 
    The best estimate of the star's location is the one that maximizes the total summed 
    energy, as determined by a 2D quadratic fit.

    All filenames may be absolute or relative paths. If relative, they will be 
    relative to the current working directory, not to any particular location in Calibration.

    Args:
        spotArray (numpy.ndarray): 
            2D array of the DM-generated satellite spots. Calculated outside 
            this function from probed images as `(Iplus + Iminus)/2 - Iunprobed`.
        xOffsetGuess (float): 
            Starting guess for the number of pixels in the x direction that the 
            star is offset from the center pixel of `spotArray`. The convention 
            for the center pixel follows that of FFTs.
        yOffsetGuess (float): 
            Starting guess for the number of pixels in the y direction that the 
            star is offset from the center pixel of `spotArray`. The convention 
            for the center pixel follows that of FFTs.
        tuningParamDict (dict): 
            Dictionary containing tuning parameter values.
            
            - **spotSepPix (float)**:  
              Expected separation of the satellite spots from the star. Used as 
              the separation for the center of the region of interest. Units of 
              pixels. Compute beforehand as separation in lambda/D multiplied by 
              pixels per lambda/D.
            - **roiRadiusPix (float)**:  
              Radius of each region of interest used when summing the intensity 
              of a satellite spot. Units of pixels.
            - **probeRotVecDeg (array_like)**:  
              1D array of angles (in degrees) counterclockwise from the x-axis 
              to rotate the regions of interest when summing the satellite spots.  
              A pair of satellite spots is given by just one value.  
              - Example for a single pair along the x-axis: `[0]` (not `[0, 180]`).  
              - Example for a plus-shaped layout of spots: `[0, 90]`.
            - **nSubpixels (int)**:  
              Number of subpixels across used to make edge values of the region-of-interest 
              mask. The value of the edge pixels in the ROI is the mean of all subpixel values.
            - **nSteps (int)**:  
              Number of points used along each direction for the grid search. 
              Odd numbers are preferable for symmetry when the array is truly centered.
            - **stepSize (float)**:  
              Step size used in the grid search. Units of pixels.
            - **nIter (int)**:  
              Number of iterations in the loop that refines the stellar center location.

    Returns:
        tuple:  
            - **xOffsetEst (float)**: Estimated lateral offset of the stellar center 
              from the center pixel of `spotArray` in the x direction.  
            - **yOffsetEst (float)**: Estimated lateral offset of the stellar center 
              from the center pixel of `spotArray` in the y direction.  
              
              The convention for the center pixel follows that of FFTs.
            - **roi_mask (numpy.ndarray)**:  
              2D float array of the best-fit region-of-interest mask used to fit 
              the translation and scaling of the ROI regions to `spotArray`.
    """

    # tuningParamDict = loadyaml(fnTuning)
    # spotSepGuessPix = tuningParamDict["spotSepGuessPix"]
    spotSepGuessPix = tuningParamDict["spotSepPix"]
    roiRadiusPix = tuningParamDict["roiRadiusPix"]
    probeRotVecDeg = tuningParamDict["probeRotVecDeg"]
    nSubpixels = tuningParamDict["nSubpixels"]
    offset_tol_pix = tuningParamDict["offset_tol_pix"]
    sep_tol_pix = tuningParamDict["sep_tol_pix"]
    opt_method = tuningParamDict["opt_method"]

    # check.real_positive_scalar(spotSepGuessPix, 'spotSepGuessPix', TypeError)
    # check.real_positive_scalar(roiRadiusPix, 'roiRadiusPix', TypeError)
    # check.oneD_array(probeRotVecDeg, 'probeRotVecDeg', TypeError)
    # check.positive_scalar_integer(nSubpixels, 'nSubpixels', TypeError)

    # Define optimization input lists
    bounds = [
        [xOffsetGuess - offset_tol_pix, xOffsetGuess + offset_tol_pix],
        [yOffsetGuess - offset_tol_pix, yOffsetGuess + offset_tol_pix],
        [spotSepGuessPix - sep_tol_pix, spotSepGuessPix + sep_tol_pix],
    ]
    initial_guess = [xOffsetGuess, yOffsetGuess, spotSepGuessPix]
    fixed_params = [spotArray, probeRotVecDeg, roiRadiusPix, nSubpixels]

    # Run the optimization
    result = minimize(
        _cost_func_spots,
        initial_guess,
        args=fixed_params,
        bounds=bounds,
        method=opt_method,
    )
    if result.success:
        fitted_params = result.x
    else:
        raise ValueError(result.message)

    param_dict = {
        "xOffset": fitted_params[0],
        "yOffset": fitted_params[1],
        "spotSepPix": fitted_params[2],
    }

    _, roi_mask = _roi_mask_for_spots(fitted_params, fixed_params)

    return param_dict, roi_mask


def calc_star_location_from_spots(spotArray, xOffsetGuess, yOffsetGuess, tuningParamDict):
    """
    Calculates the center of the occulted star using satellite spots.

    A single processed image of satellite spots is used. Multiple software masks 
    are generated and applied to the measured spots to determine the stellar location. 
    The best estimate of the star's location is the one that maximizes the total summed 
    energy, as determined by a 2D quadratic fit.

    All filenames may be absolute or relative paths. If relative, they will be 
    relative to the current working directory, not to any particular location in Calibration.

    Args:
        spotArray (numpy.ndarray): 
            2D array of the DM-generated satellite spots. Calculated outside 
            this function from probed images as `(Iplus + Iminus)/2 - Iunprobed`.
        xOffsetGuess (float): 
            Starting guess for the number of pixels in the x direction that the 
            star is offset from the center pixel of `spotArray`. The convention 
            for the center pixel follows that of FFTs.
        yOffsetGuess (float): 
            Starting guess for the number of pixels in the y direction that the 
            star is offset from the center pixel of `spotArray`. The convention 
            for the center pixel follows that of FFTs.
        tuningParamDict (dict): 
            Dictionary containing tuning parameter values.
            
            - **spotSepPix (float)**:  
              Expected separation of the satellite spots from the star. Used as 
              the separation for the center of the region of interest. Units of 
              pixels. Compute beforehand as separation in lambda/D multiplied by 
              pixels per lambda/D.
            - **roiRadiusPix (float)**:  
              Radius of each region of interest used when summing the intensity 
              of a satellite spot. Units of pixels.
            - **probeRotVecDeg (array_like)**:  
              1D array of angles (in degrees) counterclockwise from the x-axis 
              to rotate the regions of interest when summing the satellite spots.  
              A pair of satellite spots is given by just one value.  
              - Example for a single pair along the x-axis: `[0]` (not `[0, 180]`).  
              - Example for a plus-shaped layout of spots: `[0, 90]`.
            - **nSubpixels (int)**:  
              Number of subpixels across used to make edge values of the region-of-interest 
              mask. The value of the edge pixels in the ROI is the mean of all subpixel values.
            - **nSteps (int)**:  
              Number of points used along each direction for the grid search. 
              Odd numbers are preferable for symmetry when the array is truly centered.
            - **stepSize (float)**:  
              Step size used in the grid search. Units of pixels.
            - **nIter (int)**:  
              Number of iterations in the loop that refines the stellar center location.

    Returns:
        tuple:  
            - **xOffsetEst (float)**: Estimated lateral offset of the stellar center 
              from the center pixel of `spotArray` in the x direction.  
            - **yOffsetEst (float)**: Estimated lateral offset of the stellar center 
              from the center pixel of `spotArray` in the y direction.  
              
            The convention for the center pixel follows that of FFTs.
    """
    
    # check.twoD_array(spotArray, 'spotArray', TypeError)
    # check.real_scalar(xOffsetGuess, 'xOffsetGuess', TypeError)
    # check.real_scalar(yOffsetGuess, 'yOffsetGuess', TypeError)

    # tuningParamDict = loadyaml(fnYAML)
    spotSepPix = tuningParamDict["spotSepPix"]
    roiRadiusPix = tuningParamDict["roiRadiusPix"]
    probeRotVecDeg = tuningParamDict["probeRotVecDeg"]
    nSubpixels = tuningParamDict["nSubpixels"]
    nSteps = tuningParamDict["nSteps"]
    stepSize = tuningParamDict["stepSize"]
    nIter = tuningParamDict["nIter"]

    # check.real_positive_scalar(spotSepPix, 'spotSepPix', TypeError)
    # check.real_positive_scalar(roiRadiusPix, 'roiRadiusPix', TypeError)
    # check.oneD_array(probeRotVecDeg, 'probeRotVecDeg', TypeError)
    # check.positive_scalar_integer(nSubpixels, 'nSubpixels', TypeError)
    # check.positive_scalar_integer(nSteps, 'nSteps', TypeError)
    # check.real_positive_scalar(stepSize, 'stepSize', TypeError)
    # check.positive_scalar_integer(nIter, 'nIter', TypeError)

    ny, nx = spotArray.shape
    costFuncMat = np.zeros((nSteps, nSteps))
    xOffsetEst = 0
    yOffsetEst = 0

    for iter_ in range(nIter):
        xOffsetVec = (
            np.arange(nSteps) * stepSize - (nSteps - 1) / 2 * stepSize + xOffsetEst
        )
        yOffsetVec = (
            np.arange(nSteps) * stepSize - (nSteps - 1) / 2 * stepSize + yOffsetEst
        )

        for iy, yOffset in enumerate(yOffsetVec):
            for ix, xOffset in enumerate(xOffsetVec):
                # Generate mask of all ROI regions
                roi_mask = np.zeros((ny, nx))
                for iRot, rotDeg in enumerate(probeRotVecDeg):
                    rotRad = np.radians(rotDeg)
                    xProbePlusCoord = np.array(
                        [
                            np.sin(rotRad) * spotSepPix + yOffsetGuess + yOffset,
                            np.cos(rotRad) * spotSepPix + xOffsetGuess + xOffset,
                        ]
                    )
                    rotRad += np.pi
                    xProbeMinusCoord = np.array(
                        [
                            np.sin(rotRad) * spotSepPix + yOffsetGuess + yOffset,
                            np.cos(rotRad) * spotSepPix + xOffsetGuess + xOffset,
                        ]
                    )

                    roi_mask += circle(
                        nx,
                        ny,
                        roiRadiusPix,
                        xProbePlusCoord[1],
                        xProbePlusCoord[0],
                        nSubpixels=nSubpixels,
                    )
                    roi_mask += circle(
                        nx,
                        ny,
                        roiRadiusPix,
                        xProbeMinusCoord[1],
                        xProbeMinusCoord[0],
                        nSubpixels=nSubpixels,
                    )

                costFuncMat[iy, ix] = np.sum(roi_mask * spotArray)

        xOffsetEst, yOffsetEst = find_optimum_2d(
            xOffsetVec, yOffsetVec, costFuncMat, np.ones_like(costFuncMat)
        )

    # estimates are w.r.t. initial guess. Return, "from the center pixel"
    return xOffsetEst + xOffsetGuess, yOffsetEst + yOffsetGuess


def calc_spot_separation(spotArray, xOffset, yOffset, tuningParamDict):
    """
    Calculates the radial separation in pixels of the satellite spots.

    A single processed image of satellite spots is used. Several software masks 
    are generated and applied to the measured spots to determine the radial 
    separation of the spots from the given star center. The best estimate of 
    the radial spot separation is the one that maximizes the total summed 
    energy in the software mask, as determined by a 1D quadratic fit.

    All filenames may be absolute or relative paths. If relative, they will be 
    relative to the current working directory, not to any particular location 
    in Calibration.

    Args:
        spotArray (numpy.ndarray): 
            2D array of the DM-generated satellite spots. Calculated outside 
            this function from probed images as `(Iplus + Iminus)/2 - Iunprobed`.
        xOffset (float): 
            Previously estimated stellar center offset from the array's center 
            pixel. Units of pixels. The convention for the center pixel follows 
            that of FFTs.
        yOffset (float): 
            Previously estimated stellar center offset from the array's center 
            pixel. Units of pixels. The convention for the center pixel follows 
            that of FFTs.
        tuningParamDict (dict): 
            Dictionary containing tuning parameter values.
            
            - **spotSepPix (float)**:  
              Expected (model-based) separation of the satellite spots from the 
              star. Used as the starting point for the separation for the 
              center of the region of interest. Units of pixels. Compute 
              beforehand as separation in lambda/D multiplied by pixels per 
              lambda/D (e.g., `6.5 * (51.46 * 0.575 / 13)`).
            - **roiRadiusPix (float)**:  
              Radius of each region of interest used when summing the intensity 
              of a satellite spot. Units of pixels.
            - **probeRotVecDeg (array_like)**:  
              1D array of angles (in degrees) counterclockwise from the x-axis 
              to rotate the regions of interest when summing the satellite spots. 
              A pair of satellite spots is given by just one value.  
              - Example for a single pair along the x-axis: `[0]` (not `[0, 180]`).  
              - Example for a plus-shaped layout of spots: `[0, 90]`.
            - **nSubpixels (int)**:  
              Number of subpixels across used to make edge values of the region-of-interest 
              mask. The value of the edge pixels in the ROI is the mean of all the subpixel values.
            - **nSteps (int)**:  
              Number of points used along each direction for the grid search. 
              Odd numbers are preferable for symmetry when the array is truly centered.
            - **stepSize (float)**:  
              Step size used in the grid search. Units of pixels.
            - **nIter (int)**:  
              Number of iterations in the loop that refines the radial separation of the satellite spots.

    Returns:
        float: 
            Estimated radial separation of the satellite spots from the stellar center. 
            Units of pixels.
    """
    # check.twoD_array(spotArray, 'spotArray', TypeError)
    # check.real_scalar(xOffset, 'xOffset', TypeError)
    # check.real_scalar(yOffset, 'yOffset', TypeError)

    # tuningParamDict = loadyaml(fnYAML)
    # spotSepGuessPix = tuningParamDict["spotSepGuessPix"]
    spotSepGuessPix = tuningParamDict["spotSepPix"]
    roiRadiusPix = tuningParamDict["roiRadiusPix"]
    probeRotVecDeg = tuningParamDict["probeRotVecDeg"]
    nSubpixels = tuningParamDict["nSubpixels"]
    nSteps = tuningParamDict["nSteps"]
    stepSize = tuningParamDict["stepSize"]
    nIter = tuningParamDict["nIter"]

    # check.real_positive_scalar(spotSepGuessPix, 'spotSepGuessPix', TypeError)
    # check.real_positive_scalar(roiRadiusPix, 'roiRadiusPix', TypeError)
    # check.oneD_array(probeRotVecDeg, 'probeRotVecDeg', TypeError)
    # check.positive_scalar_integer(nSubpixels, 'nSubpixels', TypeError)
    # check.positive_scalar_integer(nSteps, 'nSteps', TypeError)
    # check.real_positive_scalar(stepSize, 'stepSize', TypeError)
    # check.positive_scalar_integer(nIter, 'nIter', TypeError)

    ny, nx = spotArray.shape
    costFuncVec = np.zeros((nSteps,))
    spotSepEst = spotSepGuessPix  # initialize

    for iter_ in range(nIter):
        spotSepVec = (
            np.arange(nSteps) * stepSize - (nSteps - 1) / 2 * stepSize + spotSepEst
        )

        for iSep, spotSep in enumerate(spotSepVec):
            # Generate mask of all ROI regions
            roi_mask = np.zeros((ny, nx))
            for iRot, rotDeg in enumerate(probeRotVecDeg):
                rotRad = np.radians(rotDeg)
                xProbePlusCoord = np.array(
                    [
                        np.sin(rotRad) * spotSep + yOffset,
                        np.cos(rotRad) * spotSep + xOffset,
                    ]
                )
                rotRad += np.pi
                xProbeMinusCoord = np.array(
                    [
                        np.sin(rotRad) * spotSep + yOffset,
                        np.cos(rotRad) * spotSep + xOffset,
                    ]
                )

                roi_mask += circle(
                    nx,
                    ny,
                    roiRadiusPix,
                    xProbePlusCoord[1],
                    xProbePlusCoord[0],
                    nSubpixels=nSubpixels,
                )
                roi_mask += circle(
                    nx,
                    ny,
                    roiRadiusPix,
                    xProbeMinusCoord[1],
                    xProbeMinusCoord[0],
                    nSubpixels=nSubpixels,
                )

            costFuncVec[iSep] = np.sum(roi_mask * spotArray)

        # At the first iteration, uses the maximum instead of a
        # quadratic fit in order to get a larger capture range.
        if iter_ == 0 and nIter > 1:
            bestInd = np.argmax(costFuncVec)
            spotSepEst = spotSepVec[bestInd]
        else:
            spotSepEst = find_optimum_1d(spotSepVec, costFuncVec)

    return spotSepEst


def star_center_from_satellite_spots(
    img_ref,
    img_sat_spot,
    star_coordinate_guess,
    thetaOffsetGuess,
    satellite_spot_parameters,
):
    """
    Estimates the star center and spot locations from satellite spot images and science data.

    Args:
        img_ref (numpy.ndarray):
            2D image representing a clean occulted focal-plane image with a base DM setting.
        img_sat_spot (numpy.ndarray):
            2D image representing a clean occulted focal-plane image with a relative satellite-spot DM setting added.
        star_coordinate_guess (tuple of float):
            Starting guess for the absolute (x, y) coordinate of the star (in pixels).
            The offset calculation is referenced to the image center, which is assumed
            to be (image_size // 2, image_size // 2).
        thetaOffsetGuess (float):
            Theta rotation (in degrees) of spot locations on the camera, which might be different
            from expected due to clocking error between the DM and the camera.
        satellite_spot_parameters (dict, optional):
            Dictionary containing tuning parameters for spot separation and offset estimation. The dictionary
            should have the following structure:

            offset : dict
                Parameters for estimating the offset of the star center:

                spotSepPix : float
                    Expected (model-based) separation of the satellite spots from the star.
                    Units: pixels.
                roiRadiusPix : float
                    Radius of the region of interest around each satellite spot.
                    Units: pixels.
                probeRotVecDeg : array_like
                    Angles (degrees CCW from x-axis) specifying the position of satellite spot pairs.
                nSubpixels : int
                    Number of subpixels across for calculating region-of-interest mask edges.
                nSteps : int
                    Number of points in grid search along each direction.
                stepSize : float
                    Step size for the grid search.
                    Units: pixels.
                nIter : int
                    Number of iterations refining the radial separation.

            separation : dict
                Parameters for estimating the separation of satellite spots from the star:

                spotSepPix : float
                    Expected separation between star and satellite spots.
                    Units: pixels.
                roiRadiusPix : float
                    Radius of the region of interest around each satellite spot.
                    Units: pixels.
                probeRotVecDeg : array_like
                    Angles (degrees CCW from x-axis) specifying the position of satellite spot pairs.
                nSubpixels : int
                    Number of subpixels across for calculating region-of-interest mask edges.
                nSteps : int
                    Number of points in grid search along each direction.
                stepSize : float
                    Step size for the grid search.
                    Units: pixels.
                nIter : int
                    Number of iterations refining the radial separation.

    Returns:
        numpy.ndarray:
            Estimated absolute coordinates [x, y] of the star center in the spots image.
        numpy.ndarray:
            Calculated locations of the satellite spots.
    """

    # check inputs
    img_shp = img_ref.shape
    if img_sat_spot.shape != img_shp:
        raise TypeError("Satellite spot image not the same shape as the science image.")

    # Unpack the star guess and calculate offsets relative to image center
    xGuess, yGuess = star_coordinate_guess
    img_center_x = img_sat_spot.shape[1] // 2
    img_center_y = img_sat_spot.shape[0] // 2
    xOffsetGuess = xGuess - img_center_x
    yOffsetGuess = yGuess - img_center_y

    # Subtract reference image from satellite spot image to highlight satellite spots
    img_spots = img_sat_spot - img_ref

    # Grab the relevant tuning parameters
    tuningParamDict = satellite_spot_parameters

    # estimate star location:
    # calc_star_location_from_spots should return (xOffsetEst, yOffsetEst)
    xOffsetEst, yOffsetEst = calc_star_location_from_spots(
        img_spots, xOffsetGuess, yOffsetGuess, tuningParamDict['offset']
    )

    # estimate spot separation (from the star, i.e., radius)
    spotRadiusEst = calc_spot_separation(
        img_spots, xOffsetEst, yOffsetEst, tuningParamDict['separation']
    )

    # get spot rotation vector (angles in degrees)
    probeRotVecDeg = tuningParamDict['separation']["probeRotVecDeg"]

    # calculate locations of spot pairs
    list_spots_xy = []
    for theta_deg in probeRotVecDeg:
        theta = (theta_deg + thetaOffsetGuess) * np.pi / 180.0
        # One spot
        list_spots_xy.append([
            spotRadiusEst * np.cos(theta) + xOffsetEst,
            spotRadiusEst * np.sin(theta) + yOffsetEst,
        ])
        # Opposite spot (theta + pi)
        list_spots_xy.append([
            spotRadiusEst * np.cos(theta + np.pi) + xOffsetEst,
            spotRadiusEst * np.sin(theta + np.pi) + yOffsetEst,
        ])

    # Convert estimated offsets back to absolute coordinates
    star_xy = np.array([
        xOffsetEst + img_center_x,
        yOffsetEst + img_center_y
    ], dtype='float')

    list_spots_xy = np.array(list_spots_xy, dtype='float')

    return star_xy, list_spots_xy



import os
import json
import astropy.time as time
import warnings
import corgidrp
import corgidrp.astrom
import corgidrp.bad_pixel_calibration
import corgidrp.calibrate_kgain
import corgidrp.combine
import corgidrp.data as data
import corgidrp.caldb as caldb
import corgidrp.l1_to_l2a
import corgidrp.l2a_to_l2b
import corgidrp.l2b_to_l3
import corgidrp.l3_to_l4
import corgidrp.photon_counting
import corgidrp.pump_trap_calibration
import corgidrp.calibrate_nonlin
import corgidrp.detector
import corgidrp.flat
import corgidrp.darks
import corgidrp.sorting
import corgidrp.fluxcal

all_steps = {
    "prescan_biassub" : corgidrp.l1_to_l2a.prescan_biassub,
    "detect_cosmic_rays" : corgidrp.l1_to_l2a.detect_cosmic_rays,
    "calibrate_nonlin": corgidrp.calibrate_nonlin.calibrate_nonlin,
    "correct_nonlinearity" : corgidrp.l1_to_l2a.correct_nonlinearity,
    "update_to_l2a" : corgidrp.l1_to_l2a.update_to_l2a,
    "add_photon_noise" : corgidrp.l2a_to_l2b.add_photon_noise,
    "dark_subtraction" : corgidrp.l2a_to_l2b.dark_subtraction,
    "flat_division" : corgidrp.l2a_to_l2b.flat_division,
    "frame_select" : corgidrp.l2a_to_l2b.frame_select,
    "convert_to_electrons" : corgidrp.l2a_to_l2b.convert_to_electrons,
    "em_gain_division" : corgidrp.l2a_to_l2b.em_gain_division,
    "cti_correction" : corgidrp.l2a_to_l2b.cti_correction,
    "correct_bad_pixels" : corgidrp.l2a_to_l2b.correct_bad_pixels,
    "desmear" : corgidrp.l2a_to_l2b.desmear,
    "update_to_l2b" : corgidrp.l2a_to_l2b.update_to_l2b,
    "boresight_calibration": corgidrp.astrom.boresight_calibration,
    "calibrate_trap_pump": corgidrp.pump_trap_calibration.tpump_analysis,
    "create_bad_pixel_map" : corgidrp.bad_pixel_calibration.create_bad_pixel_map,
    "calibrate_kgain" : corgidrp.calibrate_kgain.calibrate_kgain,
    "calibrate_darks" : corgidrp.darks.calibrate_darks_lsq,
    "create_onsky_flatfield" : corgidrp.flat.create_onsky_flatfield,
    "combine_subexposures" : corgidrp.combine.combine_subexposures,
    "build_trad_dark" : corgidrp.darks.build_trad_dark,
    "sort_pupilimg_frames" : corgidrp.sorting.sort_pupilimg_frames,
    "get_pc_mean" : corgidrp.photon_counting.get_pc_mean,
    "divide_by_exptime" : corgidrp.l2b_to_l3.divide_by_exptime,
    "northup" : corgidrp.l3_to_l4.northup,
    "calibrate_fluxcal_aper": corgidrp.fluxcal.calibrate_fluxcal_aper,
    "update_to_l3": corgidrp.l2b_to_l3.update_to_l3,
    "create_wcs": corgidrp.l2b_to_l3.create_wcs,
    "distortion_correction": corgidrp.l3_to_l4.distortion_correction,
    "find_star": corgidrp.l3_to_l4.find_star,
    "do_psf_subtraction": corgidrp.l3_to_l4.do_psf_subtraction,
    "update_to_l4": corgidrp.l3_to_l4.update_to_l4,
}

recipe_dir = os.path.join(os.path.dirname(__file__), "recipe_templates")

def walk_corgidrp(filelist, CPGS_XML_filepath, outputdir, template=None):
    """
    Automatically create a recipe and process the input filelist.
    Does both the `autogen_recipe` and `run_recipe` steps.

    Args:
        filelist (list of str): list of filepaths to files
        CPGS_XML_filepath (str): path to CPGS XML file for this set of files in filelist
        outputdir (str): output directory folderpath
        template (str or json): custom template. It can be one of three things
                                  * the full json object, 
                                  * a filename of a template that's already in the recipe_templates folder
                                  * a filepath to a template on disk somewhere
                                

    Returns:
        json or list: the JSON recipe (or list of JSON recipes) that was used for processing
    """
    if isinstance(template, str):
        if os.path.sep not in template:
            # this is just a template name in the recipe_templates folder
            recipe_filepath = os.path.join(recipe_dir, template)
        else:
            recipe_filepath = template
        
        template = json.load(open(recipe_filepath, 'r'))

    # generate recipe
    recipes = autogen_recipe(filelist, outputdir, template=template)

    # process recipe
    if isinstance(recipes, list):
        # if multiple recipes
        for recipe in recipes:
            run_recipe(recipe)
    else:
        # process single recipe
        run_recipe(recipes)

    return recipes

def autogen_recipe(filelist, outputdir, template=None):
    """
    Automatically creates a recipe (or recipes) by identifyng and populating a template.
    Returns a single recipe unless there are multiple recipes that should be produced.

    Args:
        filelist (list of str): list of filepaths to files
        outputdir (str): output directory folderpath
        template (json): enables passing in of custom template, if desired

    Returns:
        json list: the JSON recipe (or list of recipes) that the input filelist will be processed with
    """
    # Handle the case where filelist is empty
    if not filelist:
        print("Input filelist is empty, using default handling to create recipe.")
        first_frame = None
    else:
        # load the data to check what kind of recipe it is
        dataset = data.Dataset(filelist)
        first_frame = dataset[0]

    # if user didn't pass in template
    if template is None:
        recipe_filename = guess_template(dataset)

        # handle it as a list moving forward
        if isinstance(recipe_filename, list):
            recipe_filename_list = recipe_filename
        else:
            recipe_filename_list = [recipe_filename]

        recipe_template_list = []
        for recipe_filename in recipe_filename_list:
            # load the template recipe
            recipe_filepath = os.path.join(recipe_dir, recipe_filename)
            template = json.load(open(recipe_filepath, 'r'))
            recipe_template_list.append(template)
    else:
        # user passed in a single template
        recipe_template_list = [template]

    recipe_list = []
    for template in recipe_template_list:
        # create the personalized recipe
        recipe = template.copy()
        recipe["template"] = False

        for filename in filelist:
            recipe["inputs"].append(filename)

        recipe["outputdir"] = outputdir

        ## Populate default values
        ## This includes calibration files that need to be automatically determined
        ## This also includes the dark subtraction outputdir for synthetic darks
        this_caldb = caldb.CalDB()
        for step in recipe["steps"]:
            # by default, identify all the calibration files needed, unless jit setting is turned on
            # two cases where we should be identifying the calibration recipes now
            if "jit_calib_id" in recipe['drpconfig'] and (not recipe['drpconfig']["jit_calib_id"]):
                _fill_in_calib_files(step, this_caldb, first_frame)
            elif ("jit_calib_id" not in recipe['drpconfig']) and (not corgidrp.jit_calib_id):
                _fill_in_calib_files(step, this_caldb, first_frame)

            if step["name"].lower() == "dark_subtraction":
                if step["keywords"]["outputdir"].upper() == "AUTOMATIC":
                    step["keywords"]["outputdir"] = recipe["outputdir"]

        recipe_list.append(recipe)
    
    # if only a single recipe, return the recipe. otherwise return list
    if len(recipe_list) > 1:
        return recipe_list
    else:
        return recipe_list[0]

def _fill_in_calib_files(step, this_caldb, ref_frame):
    """
    Fills in calibration files defined as "AUTOMATIC" in a recipe
    
    By default, throws an error if there are no available cal files of a certian type.
    Exceptional case is when the pipeline setting `skip_missing_cal_steps = True` is set:
    in this case, it will mark this step to be skipped, but continue processing the recipe.

    Args:
        step (dict): the portion of a recipe for this step
        this_caldb (corgidrp.CalDB): calibration database conection
        ref_frame (corgidrp.Image): a reference frame to use to determine the optimal calibration
    
    Returns:
        dict: the step, but with calibration files filled in
    """
    if "calibs" not in step:
        return step # don't have to do anything if no calibrations
    
    for calib in step["calibs"]:
        # order matters, so only one calibration file per dictionary

        if "AUTOMATIC" in step["calibs"][calib].upper():
            calib_dtype = data.datatypes[calib]

            # try to look up the best calibration, but it could raise an error
            try:
                best_cal_file = this_caldb.get_calib(ref_frame, calib_dtype)
                best_cal_filepath = best_cal_file.filepath
            except ValueError as e:
                if "OPTIONAL" in step["calibs"][calib].upper():
                    # couldn't find a good cal but this one is optional, so we are going to put nothing in there
                    # this means the step function can run without this calibration file
                    best_cal_filepath = None
                elif corgidrp.skip_missing_cal_steps:
                    step["skip"] = True # skip this step but continue
                    step["calibs"][calib] = None
                    warnings.warn("Skipping {0} because no {1} in caldb and skip_missing_cal_steps is True".format(step['name'], calib))
                    continue # continue on the for loop
                else:
                    raise # reraise exception

            # set calibration file to this one
            step["calibs"][calib] = best_cal_filepath

    return step

def guess_template(dataset):
    """
    Guesses what template should be used to process a specific image

    Args:
        dataset (corgidrp.data.Dataset): a Dataset to process

    Returns:
        str or list: the best template filename or a list of multiple template filenames
    """
    image = dataset[0] # first image for convenience

    # L1 -> L2a data processing
    if image.ext_hdr['DATALVL'] == "L1":
        if 'VISTYPE' not in image.pri_hdr:
            # this is probably IIT test data. Do generic processing
            recipe_filename = "l1_to_l2b.json"
        elif image.pri_hdr['VISTYPE'][:3] == "ENG":
            # first three letters are ENG
            # for either ENGPUPIL or ENGIMGAGE
            recipe_filename = "l1_to_l2a_eng.json"
        elif image.pri_hdr['VISTYPE'] == "BORESITE":
            recipe_filename = "l1_to_boresight.json"
        elif image.pri_hdr['VISTYPE'] == "FFIELD":
            recipe_filename = "l1_flat_and_bp.json"
        elif image.pri_hdr['VISTYPE'] == "DARK":
            _, unique_vals = dataset.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
            if image.ext_hdr['ISPC']:
                recipe_filename = "l1_to_l2b_pc_dark.json"
            elif len(unique_vals) > 1: # darks for noisemap creation
                recipe_filename = "l1_to_l2a_noisemap.json"
            else: # then len(unique_vals) is 1 and not PC: traditional darks
                recipe_filename = "build_trad_dark_image.json"
        elif image.pri_hdr['VISTYPE'] == "PUPILIMG":
            recipe_filename = ["l1_to_l2a_nonlin.json", "l1_to_kgain.json"]
        else:
            recipe_filename = "l1_to_l2a_basic.json"  # science data and all else (including photon counting)
    # L2a -> L2b data processing
    elif image.ext_hdr['DATALVL'] == "L2a":
        if image.pri_hdr['VISTYPE'] == "DARK":
            _, unique_vals = dataset.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
            if image.ext_hdr['ISPC']:
                recipe_filename = "l2a_to_l2b_pc_dark.json"
            elif len(unique_vals) > 1: # darks for noisemap creation
                recipe_filename = "l2a_to_l2a_noisemap.json"
            else: # then len(unique_vals) is 1 and not PC: traditional darks
                recipe_filename = "l2a_build_trad_dark_image.json"
        else:
            if image.ext_hdr['ISPC']:
                recipe_filename = "l2a_to_l2b_pc.json"
            else:
                recipe_filename = "l2a_to_l2b.json"  # science data and all else
    # L2b -> L3 data processing
    elif image.ext_hdr['DATALVL'] == "L2b":
        if image.pri_hdr['VISTYPE'] == "ABSFLXFT" or image.pri_hdr['VISTYPE'] == "ABSFLXBT":
            recipe_filename = "l2b_to_fluxcal_factor.json"
        else:
            recipe_filename = "l2b_to_l3.json"
    # L3 -> L4 data processing
    elif image.ext_hdr['DATALVL'] == "L3":
        if image.ext_hdr['FSMLOS'] == 1:
            # coronagraphic obs - PSF subtraction
            recipe_filename = "l3_to_l4.json"
        else:
            # noncorongrpahic obs - no PSF subtraction
            recipe_filename = "l3_to_l4_nopsfsub.json"
            
    else:
        raise NotImplementedError("Cannot automatically guess the input dataset with 'DATALVL' = {0}".format(image.ext_hdr['DATALVL']))

    return recipe_filename


def save_data(dataset_or_image, outputdir, suffix=""):
    """
    Saves the dataset or image that has currently been outputted by the last step function.
    Records calibration frames into the caldb during the process

    Args:
        dataset_or_image (corgidrp.data.Dataset or corgidrp.data.Image): data to save
        outputdir (str): path to directory where files should be saved
        suffix (str): optional suffix to tack onto the filename. 
                      E.g.: `test.fits` with `suffix="dark"` becomes `test_dark.fits`
    """
    # convert everything to dataset to make life easier
    if isinstance(dataset_or_image, data.Image):
        dataset = data.Dataset([dataset_or_image])
    else:
        dataset = dataset_or_image

    # add suffix to ending if necessary
    if len(suffix) > 0:
        filenames = []

        suffix = suffix.strip("_") # user doesn't need to pass underscores
        for image in dataset:
            # grab everything before .FITS
            fits_index = image.filename.lower().rfind(".fits")
            filename_base = image.filename[:fits_index]
            new_filename = "{0}_{1}.fits".format(filename_base, suffix)
            filenames.append(new_filename)
    else:
        filenames = None

    # save!
    dataset.save(filedir=outputdir, filenames=filenames)

    # add calibration data to caldb as necessary
    for image in dataset:
        if type(image) in caldb.labels:
            # this is a calibration frame!
            this_caldb = caldb.CalDB()
            this_caldb.create_entry(image)



def run_recipe(recipe, save_recipe_file=True):
    """
    Run the specified recipe

    Args:
        recipe (dict or str): either the filepath to the recipe or the already loaded in recipe
        save_recipe_file (bool): saves the recipe as a JSON file in the outputdir (true by default)
    """
    if isinstance(recipe, str):
        # need to load in
        recipe = json.load(open(recipe, "r"))

    # configure pipeline as needed
    for setting in recipe['drpconfig']:
        # equivalent to corgidrp.setting = recipe['drpconfig'][setting]
        setattr(corgidrp, setting, recipe['drpconfig'][setting])

    # read in data, if not doing bp map
    if recipe["inputs"]:
        filelist = recipe["inputs"]
        curr_dataset = data.Dataset(filelist)
        # write the recipe into the image extension header
        for frame in curr_dataset:
            frame.ext_hdr["RECIPE"] = json.dumps(recipe)
    else:
        curr_dataset = []

    # save recipe before running recipe
    if save_recipe_file:
        recipe_filename = "{0}_{1}_recipe.json".format(recipe["name"], time.Time.now().isot)
        recipe_filepath = os.path.join(recipe["outputdir"], recipe_filename)
        with open(recipe_filepath, "w") as json_file:
            json.dump(recipe, json_file, indent=4)

    tot_steps = len(recipe["steps"])

    # execute each pipeline step
    for i, step in enumerate(recipe["steps"]):
        print("Walker step {0}/{1}: {2}".format(i+1, tot_steps, step["name"]))
        if step["name"].lower() == "save":
            # special save instruction
            
            # see if suffix is specified as a keyword
            if "keywords" in step and "suffix" in step["keywords"]:
                suffix =  step["keywords"]["suffix"]
            else:
                suffix = ''
                
            save_data(curr_dataset, recipe["outputdir"], suffix=suffix)

        else:
            step_func = all_steps[step["name"]]

            # edge case if this step has been specified to be skipped
            if "skip" in step and step["skip"]:
                continue

            other_args = ()
            if "calibs" in step:
                # if JIT calibration resolving is toggled, figure out the calibrations here
                # by default, this is false
                if (corgidrp.jit_calib_id and ("jit_calib_id" not in recipe['drpconfig'])) or (("jit_calib_id" in recipe['drpconfig']) and recipe['drpconfig']["jit_calib_id"]) :
                    this_caldb = caldb.CalDB()
                    # dataset may have turned into a single image. handle this case. 
                    if isinstance(curr_dataset, data.Dataset):
                        ref_image = curr_dataset[0]
                        list_of_frames = curr_dataset
                    else:
                        ref_image = curr_dataset
                        list_of_frames = [curr_dataset]
                    _fill_in_calib_files(step, this_caldb, ref_image)

                    # also update the recipe we used in the headers
                    for frame in list_of_frames:
                        frame.ext_hdr["RECIPE"] = json.dumps(recipe)


                # load the calibration files in from disk
                for calib in step["calibs"]:
                    calib_dtype = data.datatypes[calib]
                    if step["calibs"][calib] is not None:
                        cal_file = calib_dtype(step["calibs"][calib])
                    else:
                        cal_file = None
                    other_args += (cal_file,)


            if "keywords" in step:
                kwargs = step["keywords"]
            else:
                kwargs = {}

            # run the step!
            curr_dataset = step_func(curr_dataset, *other_args, **kwargs)




import numpy as np
import os
from corgidrp import mocks, astrom
from corgidrp.l2b_to_l3 import create_wcs

def test_create_wcs():
    """
    Unit test of the create WCS function. The test checks that all WCS keywords (['CD{}_{}'], ['CRPIX{}'], ['CTYPE{}'], ['CDELT{}'], ['CRVAL{}']) exist in the updated ext_hdr 
    and that the values are consistent with those derived from the AstrometricCalibration file.
    """

    # create mock dataset (arbitrary northangle)
    field_path = os.path.join(os.path.dirname(__file__), "test_data", "JWST_CALFIELD2020.csv")
    mock_dataset = mocks.create_astrom_data(field_path, platescale=21.8, rotation=20, dither_pointings=2)

    # run the boresight calibration to get an AstrometricCalibration file
    astrom_cal = astrom.boresight_calibration(mock_dataset, field_path, find_threshold=200, find_distortion=True, position_error=0.5)

    # create the wcs
    updated_dataset = create_wcs(mock_dataset, astrom_cal)

    # check that all wcs keywords exist in the ext_hdrs of the mock dataset
    # and that the values are as expected from the AstrometricCalibration file
    platescale_x = astrom_cal.platescale[0]
    platescale_y = astrom_cal.platescale[1]
    northangle = astrom_cal.northangle
    ra_offset, dec_offset = astrom_cal.avg_offset

    for mock_frame, updated_frame in zip(mock_dataset, updated_dataset):
        roll_ang = mock_frame.pri_hdr['ROLL']
        data = mock_frame.data
        image_shape = data.shape
        center_pixel = [(image_shape[1]-1) // 2, (image_shape[0]-1) // 2]
        target_ra, target_dec = mock_frame.pri_hdr['RA'], mock_frame.pri_hdr['DEC']

        pc = np.array([[-np.cos(np.radians(northangle + roll_ang)), np.sin(np.radians(northangle + roll_ang))], [np.sin(np.radians(northangle + roll_ang)), np.cos(np.radians(northangle + roll_ang))]])
        # assuming platescale is the same along both axes
        matrix = pc * (platescale_x * 0.001) / 3600.
        
        # gather expected values in a dictionary
        expected = {}
        expected['CD1_1'] = matrix[0,0]
        expected['CD1_2'] = matrix[0,1]
        expected['CD2_1'] = matrix[1,0]
        expected['CD2_2'] = matrix[1,1]

        expected['CRPIX1'] = center_pixel[0]
        expected['CRPIX2'] = center_pixel[1]

        expected['CTYPE1'] = 'RA---TAN'
        expected['CTYPE2'] = 'DEC--TAN'

        expected['CDELT1'] = (platescale_x * 0.001) / 3600  ## converting to degrees
        expected['CDELT2'] = (platescale_y * 0.001) / 3600

        expected['CRVAL1'] = target_ra - ra_offset      # the corrected target pointing based on astrom_cal
        expected['CRVAL2'] = target_dec - dec_offset

        # gather the wcs values from the updated dataset
        wcs = {}
        for key in expected.keys():
            wcs[key] = updated_frame.ext_hdr[key]

        # compare the expected dictionary to the updated dateset output
        assert wcs.items() == expected.items()
if __name__ == "__main__":
    test_create_wcs()